# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import errno
import logging
import os
import os.path
import shutil
import subprocess
import sys
import tempfile

import ansible.constants

from kayobe import exception
from kayobe import utils
from kayobe import vault


DEFAULT_CONFIG_PATH = "/etc/kayobe"

CONFIG_PATH_ENV = "KAYOBE_CONFIG_PATH"

ENVIRONMENT_ENV = "KAYOBE_ENVIRONMENT"

LOG = logging.getLogger(__name__)


def add_args(parser):
    """Add arguments required for running Ansible playbooks to a parser."""
    default_config_path = os.getenv(CONFIG_PATH_ENV, DEFAULT_CONFIG_PATH)
    default_environment = os.getenv(ENVIRONMENT_ENV)
    parser.add_argument("-b", "--become", action="store_true",
                        help="run operations with become (nopasswd implied)")
    parser.add_argument("-C", "--check", action="store_true",
                        help="don't make any changes; instead, try to predict "
                             "some of the changes that may occur")
    parser.add_argument("--config-path", default=default_config_path,
                        help="path to Kayobe configuration. "
                             "(default=$%s or %s)" %
                             (CONFIG_PATH_ENV, DEFAULT_CONFIG_PATH))
    parser.add_argument("-D", "--diff", action="store_true",
                        help="when changing (small) files and templates, show "
                             "the differences in those files; works great "
                             "with --check")
    parser.add_argument("--environment", default=default_environment,
                        help="specify environment name (default=$%s or None)" %
                             ENVIRONMENT_ENV)
    parser.add_argument("-e", "--extra-vars", metavar="EXTRA_VARS",
                        action="append",
                        help="set additional variables as key=value or "
                             "YAML/JSON")
    parser.add_argument("-i", "--inventory", metavar="INVENTORY",
                        action="append",
                        help="specify inventory host path "
                             "(default=$%s/inventory or %s/inventory) " %
                             (CONFIG_PATH_ENV, DEFAULT_CONFIG_PATH))
    parser.add_argument("-l", "--limit", metavar="SUBSET",
                        help="further limit selected hosts to an additional "
                             "pattern")
    parser.add_argument("--skip-tags", metavar="TAGS",
                        help="only run plays and tasks whose tags do not "
                             "match these values")
    parser.add_argument("-t", "--tags", metavar="TAGS",
                        help="only run plays and tasks tagged with these "
                             "values")
    parser.add_argument("-lt", "--list-tasks",
                        action="store_true",
                        help="only print names of tasks, don't run them, "
                             "note this has no affect on kolla-ansible.")
    parser.add_argument("-sh", "--skip-hooks", action="store", default=None,
                        help="disables hooks. Specify a pattern to skip"
                             "specific playbooks. \"all\" skips all playbooks")


def _get_kayobe_environment_path(parsed_args):
    """Return the path to the Kayobe environment or None if not specified."""
    env_path = None
    if parsed_args.environment:
        # Specified via --environment or KAYOBE_ENVIRONMENT.
        kc_environments = os.path.join(parsed_args.config_path, "environments")
        env_path = os.path.join(kc_environments, parsed_args.environment)
    return env_path


def _get_inventories_paths(parsed_args, env_path):
    """Return the paths to the Kayobe inventories."""
    default_inventory = utils.get_data_files_path("ansible", "inventory")
    inventories = [default_inventory]
    if parsed_args.inventory:
        inventories.extend(parsed_args.inventory)
    else:
        shared_inventory = os.path.join(parsed_args.config_path, "inventory")
        if env_path:
            if os.path.exists(shared_inventory):
                inventories.append(shared_inventory)
            env_inventory = os.path.join(env_path, "inventory")
            if os.path.exists(env_inventory):
                inventories.append(env_inventory)
        else:
            # Preserve existing behaviour: don't check if an inventory
            # directory exists when no environment is specified
            inventories.append(shared_inventory)
    return inventories


def _validate_args(parsed_args, playbooks):
    """Validate Kayobe Ansible arguments."""
    vault.enforce_single_password_source(parsed_args)
    result = utils.is_readable_dir(parsed_args.config_path)
    if not result["result"]:
        LOG.error("Kayobe configuration path %s is invalid: %s",
                  parsed_args.config_path, result["message"])
        sys.exit(1)

    env_path = _get_kayobe_environment_path(parsed_args)
    if env_path:
        result = utils.is_readable_dir(env_path)
        if not result["result"]:
            LOG.error("Kayobe environment %s is invalid: %s",
                      env_path, result["message"])
            sys.exit(1)

    inventories = _get_inventories_paths(parsed_args, env_path)
    for inventory in inventories:
        result = utils.is_readable_dir(inventory)
        if not result["result"]:
            LOG.error("Kayobe inventory %s is invalid: %s",
                      inventory, result["message"])
            sys.exit(1)

    for playbook in playbooks:
        result = utils.is_readable_file(playbook)
        if not result["result"]:
            LOG.error("Kayobe playbook %s is invalid: %s",
                      playbook, result["message"])
            sys.exit(1)


def _get_vars_files(vars_paths):
    """Return a list of Kayobe Ansible configuration variable files.

    The list of directories given as argument is searched to create the list of
    variable files. The files will be sorted alphabetically by name for each
    directory, but ordering of directories is kept to allow overrides.
    """
    vars_files = []
    for vars_path in vars_paths:
        path_vars_files = []
        for vars_file in os.listdir(vars_path):
            abs_path = os.path.join(vars_path, vars_file)
            if utils.is_readable_file(abs_path)["result"]:
                root, ext = os.path.splitext(vars_file)
                if ext in (".yml", ".yaml", ".json"):
                    path_vars_files.append(abs_path)
        vars_files += sorted(path_vars_files)

    return vars_files


def build_args(parsed_args, playbooks,
               extra_vars=None, limit=None, tags=None, verbose_level=None,
               check=None, ignore_limit=False, list_tasks=None, diff=None):
    """Build arguments required for running Ansible playbooks."""
    cmd = ["ansible-playbook"]
    if verbose_level:
        cmd += ["-" + "v" * verbose_level]
    if list_tasks or (parsed_args.list_tasks and list_tasks is None):
        cmd += ["--list-tasks"]
    cmd += vault.build_args(parsed_args, "--vault-password-file")
    env_path = _get_kayobe_environment_path(parsed_args)
    inventories = _get_inventories_paths(parsed_args, env_path)
    for inventory in inventories:
        cmd += ["--inventory", inventory]
    vars_paths = [parsed_args.config_path]
    if env_path:
        vars_paths.append(env_path)
    vars_files = _get_vars_files(vars_paths)
    for vars_file in vars_files:
        cmd += ["-e", "@%s" % vars_file]
    if parsed_args.extra_vars:
        for extra_var in parsed_args.extra_vars:
            # Don't quote or escape variables passed via the kayobe -e CLI
            # argument, to match Ansible's behaviour.
            cmd += ["-e", extra_var]
    if extra_vars:
        for extra_var_name, extra_var_value in extra_vars.items():
            # Quote and escape variables originating within the python CLI.
            extra_var_value = utils.quote_and_escape(extra_var_value)
            cmd += ["-e", "%s=%s" % (extra_var_name, extra_var_value)]
    if parsed_args.become:
        cmd += ["--become"]
    if check or (parsed_args.check and check is None):
        cmd += ["--check"]
    if diff or (parsed_args.diff and diff is None):
        cmd += ["--diff"]
    if not ignore_limit and (parsed_args.limit or limit):
        limit_arg = utils.intersect_limits(parsed_args.limit, limit)
        cmd += ["--limit", limit_arg]
    if parsed_args.skip_tags:
        cmd += ["--skip-tags", parsed_args.skip_tags]
    if parsed_args.tags or tags:
        all_tags = [t for t in [parsed_args.tags, tags] if t]
        cmd += ["--tags", ",".join(all_tags)]
    cmd += playbooks
    return cmd


def _get_environment(parsed_args):
    """Return an environment dict for executing an Ansible playbook."""
    env = os.environ.copy()
    vault.update_environment(parsed_args, env)
    # If the configuration path has been specified via --config-path, ensure
    # the environment variable is set, so that it can be referenced by
    # playbooks.
    env.setdefault(CONFIG_PATH_ENV, parsed_args.config_path)
    # If an environment has been specified via --environment, ensure the
    # environment variable is set, so that it can be referenced by playbooks.
    if parsed_args.environment:
        env.setdefault(ENVIRONMENT_ENV, parsed_args.environment)
    # If a custom Ansible configuration file exists, use it.
    ansible_cfg_path = os.path.join(parsed_args.config_path, "ansible.cfg")
    if utils.is_readable_file(ansible_cfg_path)["result"]:
        env.setdefault("ANSIBLE_CONFIG", ansible_cfg_path)

    # Update various role, collection and plugin paths to include the Kayobe
    # roles, collections and plugins. This allows custom playbooks to use these
    # resources.
    roles_paths = [
        os.path.join(parsed_args.config_path, "ansible", "roles"),
        utils.get_data_files_path("ansible", "roles"),
    ] + ansible.constants.DEFAULT_ROLES_PATH
    env.setdefault("ANSIBLE_ROLES_PATH", ":".join(roles_paths))

    collections_paths = [
        os.path.join(parsed_args.config_path, "ansible", "collections"),
        utils.get_data_files_path("ansible", "collections"),
    ] + ansible.constants.COLLECTIONS_PATHS
    env.setdefault("ANSIBLE_COLLECTIONS_PATH", ":".join(collections_paths))

    action_plugins = [
        os.path.join(parsed_args.config_path, "ansible", "action_plugins"),
        utils.get_data_files_path("ansible", "action_plugins"),
    ] + ansible.constants.DEFAULT_ACTION_PLUGIN_PATH
    env.setdefault("ANSIBLE_ACTION_PLUGINS", ":".join(action_plugins))

    filter_plugins = [
        os.path.join(parsed_args.config_path, "ansible", "filter_plugins"),
        utils.get_data_files_path("ansible", "filter_plugins"),
    ] + ansible.constants.DEFAULT_FILTER_PLUGIN_PATH
    env.setdefault("ANSIBLE_FILTER_PLUGINS", ":".join(filter_plugins))

    test_plugins = [
        os.path.join(parsed_args.config_path, "ansible", "test_plugins"),
        utils.get_data_files_path("ansible", "test_plugins"),
    ] + ansible.constants.DEFAULT_TEST_PLUGIN_PATH
    env.setdefault("ANSIBLE_TEST_PLUGINS", ":".join(test_plugins))

    return env


def run_playbooks(parsed_args, playbooks,
                  extra_vars=None, limit=None, tags=None, quiet=False,
                  check_output=False, verbose_level=None, check=None,
                  ignore_limit=False, list_tasks=None, diff=None):
    """Run a Kayobe Ansible playbook."""
    _validate_args(parsed_args, playbooks)
    cmd = build_args(parsed_args, playbooks,
                     extra_vars=extra_vars, limit=limit, tags=tags,
                     verbose_level=verbose_level, check=check,
                     ignore_limit=ignore_limit, list_tasks=list_tasks,
                     diff=diff)
    env = _get_environment(parsed_args)
    try:
        utils.run_command(cmd, check_output=check_output, quiet=quiet, env=env)
    except subprocess.CalledProcessError as e:
        LOG.error("Kayobe playbook(s) %s exited %d",
                  ", ".join(playbooks), e.returncode)
        if check_output:
            LOG.error("The output was:\n%s", e.output)
        sys.exit(e.returncode)


def run_playbook(parsed_args, playbook, *args, **kwargs):
    """Run a Kayobe Ansible playbook."""
    return run_playbooks(parsed_args, [playbook], *args, **kwargs)


def config_dump(parsed_args, host=None, hosts=None, var_name=None,
                facts=None, extra_vars=None, tags=None, verbose_level=None):
    dump_dir = tempfile.mkdtemp()
    try:
        if not extra_vars:
            extra_vars = {}
        extra_vars["dump_path"] = dump_dir
        if host or hosts:
            extra_vars["dump_hosts"] = host or hosts
        if var_name:
            extra_vars["dump_var_name"] = var_name
        if facts is not None:
            extra_vars["dump_facts"] = facts
        # Don't use check mode or list tasks for configuration dumps as we
        # won't get any results back.
        playbook_path = utils.get_data_files_path("ansible", "dump-config.yml")
        run_playbook(parsed_args, playbook_path,
                     extra_vars=extra_vars, tags=tags, check_output=True,
                     verbose_level=verbose_level, check=False,
                     list_tasks=False, diff=False)
        hostvars = {}
        for path in os.listdir(dump_dir):
            LOG.debug("Found dump file %s", path)
            inventory_hostname, ext = os.path.splitext(path)
            if ext == ".yml":
                hvars = utils.read_yaml_file(os.path.join(dump_dir, path))
                if host:
                    return hvars
                else:
                    hostvars[inventory_hostname] = hvars
            else:
                LOG.warning("Unexpected extension on config dump file %s",
                            path)
        return hostvars
    finally:
        shutil.rmtree(dump_dir)


def install_galaxy_roles(parsed_args, force=False):
    """Install Ansible Galaxy role dependencies.

    Installs role dependencies specified in kayobe, and if present, in kayobe
    configuration.

    :param parsed_args: Parsed command line arguments.
    :param force: Whether to force reinstallation of roles.
    """
    LOG.info("Installing galaxy role dependencies from kayobe")
    requirements = utils.get_data_files_path("requirements.yml")
    roles_destination = utils.get_data_files_path('ansible', 'roles')
    utils.galaxy_role_install(requirements, roles_destination, force=force)

    # Check for requirements in kayobe configuration.
    kc_reqs_path = os.path.join(parsed_args.config_path,
                                "ansible", "requirements.yml")
    if not utils.is_readable_file(kc_reqs_path)["result"]:
        LOG.info("Not installing galaxy role dependencies from kayobe config "
                 "- requirements.yml not present")
        return

    LOG.info("Installing galaxy role dependencies from kayobe config")
    # Ensure a roles directory exists in kayobe-config.
    kc_roles_path = os.path.join(parsed_args.config_path,
                                 "ansible", "roles")
    try:
        os.makedirs(kc_roles_path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise exception.Error("Failed to create directory ansible/roles/ "
                                  "in kayobe configuration at %s: %s" %
                                  (parsed_args.config_path, str(e)))

    # Install roles from kayobe-config.
    utils.galaxy_role_install(kc_reqs_path, kc_roles_path, force=force)


def install_galaxy_collections(parsed_args, force=False):
    """Install Ansible Galaxy collection dependencies.

    Installs collection dependencies specified in kayobe, and if present, in
    kayobe configuration.

    :param parsed_args: Parsed command line arguments.
    :param force: Whether to force reinstallation of roles.
    """
    LOG.info("Installing galaxy collection dependencies from kayobe")
    requirements = utils.get_data_files_path("requirements.yml")
    collections_destination = utils.get_data_files_path('ansible',
                                                        'collections')
    utils.galaxy_collection_install(requirements, collections_destination,
                                    force=force)

    # Check for requirements in kayobe configuration.
    kc_reqs_path = os.path.join(parsed_args.config_path,
                                "ansible", "requirements.yml")
    if not utils.is_readable_file(kc_reqs_path)["result"]:
        LOG.info("Not installing galaxy collection dependencies from kayobe "
                 "config - requirements.yml not present")
        return

    LOG.info("Installing galaxy collection dependencies from kayobe config")
    # Ensure a collections directory exists in kayobe-config.
    kc_collections_path = os.path.join(parsed_args.config_path,
                                       "ansible", "collections")
    try:
        os.makedirs(kc_collections_path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise exception.Error("Failed to create directory "
                                  "ansible/collections/ "
                                  "in kayobe configuration at %s: %s" %
                                  (parsed_args.config_path, str(e)))

    # Install collections from kayobe-config.
    utils.galaxy_collection_install(kc_reqs_path, kc_collections_path,
                                    force=force)


def prune_galaxy_roles(parsed_args):
    """Prune galaxy roles that are no longer necessary.

    :param parsed_args: Parsed command line arguments.
    """
    LOG.info("Removing unnecessary galaxy roles from kayobe")
    roles_to_remove = [
        'resmo.ntp',
        'stackhpc.ntp',
        'stackhpc.os-shade',
        'yatesr.timezone',
    ]
    LOG.debug("Removing roles: %s", ",".join(roles_to_remove))
    utils.galaxy_remove(roles_to_remove, "ansible/roles")


def passwords_yml_exists(parsed_args):
    """Return whether passwords.yml exists in the kayobe configuration."""
    env_path = _get_kayobe_environment_path(parsed_args)
    path = env_path if env_path else parsed_args.config_path
    passwords_path = os.path.join(path, 'kolla', 'passwords.yml')
    return utils.is_readable_file(passwords_path)["result"]



# Copyright (c) 2018 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import os
import unittest
from unittest import mock

from kayobe import utils
from kayobe import vault


class TestCase(unittest.TestCase):

    maxDiff = None

    @mock.patch.object(vault.utils, "run_command", autospec=True)
    def test__get_vault_password_helper(self, mock_run):
        mock_run.return_value = "fake-password\n"
        result = vault._get_vault_password_helper()
        mock_run.assert_called_once_with(
            ["which", "kayobe-vault-password-helper"], check_output=True,
            universal_newlines=True)
        self.assertEqual('fake-password', result)

    def test_enforce_single_password_source_ok(self):
        parser = argparse.ArgumentParser()
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        vault.enforce_single_password_source(parsed_args)

    @mock.patch.dict(os.environ, {"KAYOBE_VAULT_PASSWORD": "test-pass"})
    def test_enforce_single_password_source_env_kayobe(self):
        parser = argparse.ArgumentParser()
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        vault.enforce_single_password_source(parsed_args)

    @mock.patch.dict(os.environ, {"ANSIBLE_VAULT_PASSWORD_FILE":
                                  "/path/to/file"})
    def test_enforce_single_password_source_env_ansible(self):
        parser = argparse.ArgumentParser()
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        vault.enforce_single_password_source(parsed_args)

    @mock.patch.dict(os.environ, {"KAYOBE_VAULT_PASSWORD": "test-pass"})
    def test_enforce_single_password_source_ask_vault_pass(self):
        parser = argparse.ArgumentParser()
        vault.add_args(parser)
        parsed_args = parser.parse_args(["--ask-vault-pass"])
        self.assertRaises(SystemExit, vault.enforce_single_password_source,
                          parsed_args)

    @mock.patch.dict(os.environ, {"KAYOBE_VAULT_PASSWORD": "test-pass"})
    def test_enforce_single_password_source_vault_password_file(self):
        parser = argparse.ArgumentParser()
        vault.add_args(parser)
        parsed_args = parser.parse_args(["--vault-password-file",
                                         "/path/to/file"])
        self.assertRaises(SystemExit, vault.enforce_single_password_source,
                          parsed_args)

    @mock.patch.dict(os.environ, {"KAYOBE_VAULT_PASSWORD": "test-pass",
                                  "ANSIBLE_VAULT_PASSWORD_FILE":
                                  "/path/to/file"})
    def test_enforce_single_password_source_vault_both_env(self):
        parser = argparse.ArgumentParser()
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        self.assertRaises(SystemExit, vault.enforce_single_password_source,
                          parsed_args)

    @mock.patch.object(vault.getpass, 'getpass')
    def test__ask_vault_pass(self, mock_getpass):
        mock_getpass.return_value = 'test-pass'

        # Call twice to verify that the user is only prompted once.
        result = vault._ask_vault_pass()
        self.assertEqual('test-pass', result)
        mock_getpass.assert_called_once_with("Vault password: ")

        result = vault._ask_vault_pass()
        self.assertEqual('test-pass', result)
        mock_getpass.assert_called_once_with("Vault password: ")

    @mock.patch.object(utils, 'read_file')
    def test__read_vault_password_file(self, mock_read):
        mock_read.return_value = "test-pass\n"
        result = vault._read_vault_password_file("/path/to/file")
        self.assertEqual("test-pass", result)
        mock_read.assert_called_once_with("/path/to/file")

    def test_update_environment_no_vault(self):
        parser = argparse.ArgumentParser()
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        env = {}
        vault.update_environment(parsed_args, env)
        self.assertEqual({}, env)

    @mock.patch.object(vault, '_ask_vault_pass')
    def test_update_environment_prompt(self, mock_ask):
        mock_ask.return_value = "test-pass"
        parser = argparse.ArgumentParser()
        vault.add_args(parser)
        parsed_args = parser.parse_args(["--ask-vault-pass"])
        env = {}
        vault.update_environment(parsed_args, env)
        self.assertEqual({"KAYOBE_VAULT_PASSWORD": "test-pass"}, env)
        mock_ask.assert_called_once_with()

    @mock.patch.object(vault, '_read_vault_password_file')
    def test_update_environment_file_arg(self, mock_read):
        mock_read.return_value = "test-pass"
        parser = argparse.ArgumentParser()
        vault.add_args(parser)
        args = ["--vault-password-file", "/path/to/file"]
        parsed_args = parser.parse_args(args)
        env = {}
        vault.update_environment(parsed_args, env)
        self.assertEqual({"KAYOBE_VAULT_PASSWORD": "test-pass"}, env)
        mock_read.assert_called_once_with("/path/to/file")

    @mock.patch.dict(os.environ, {"ANSIBLE_VAULT_PASSWORD_FILE":
                                  "/path/to/file"})
    @mock.patch.object(vault, '_read_vault_password_file')
    def test_update_environment_file_env(self, mock_read):
        mock_read.return_value = "test-pass"
        parser = argparse.ArgumentParser()
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        env = {"ANSIBLE_VAULT_PASSWORD_FILE": "/path/to/file"}
        vault.update_environment(parsed_args, env)
        self.assertEqual({"KAYOBE_VAULT_PASSWORD": "test-pass",
                          "ANSIBLE_VAULT_PASSWORD_FILE": "/path/to/file"}, env)
        mock_read.assert_called_once_with("/path/to/file")



# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import os.path
import subprocess
import sys

from kayobe import utils
from kayobe import vault


DEFAULT_CONFIG_PATH = "/etc/kolla"

CONFIG_PATH_ENV = "KOLLA_CONFIG_PATH"

DEFAULT_VENV_PATH = "venvs/kolla-ansible"

VENV_PATH_ENV = "KOLLA_VENV_PATH"

LOG = logging.getLogger(__name__)


def add_args(parser):
    """Add arguments required for running Kolla Ansible to a parser."""
    # $KOLLA_CONFIG_PATH or /etc/kolla.
    default_config_path = os.getenv(CONFIG_PATH_ENV, DEFAULT_CONFIG_PATH)
    # $KOLLA_VENV_PATH or $PWD/venvs/kolla-ansible
    default_venv = os.getenv(VENV_PATH_ENV,
                             os.path.join(os.getcwd(), DEFAULT_VENV_PATH))
    parser.add_argument("--kolla-config-path", default=default_config_path,
                        help="path to Kolla configuration. "
                             "(default=$%s or %s)" %
                             (CONFIG_PATH_ENV, DEFAULT_CONFIG_PATH))
    parser.add_argument("-ke", "--kolla-extra-vars", metavar="EXTRA_VARS",
                        action="append",
                        help="set additional variables as key=value or "
                             "YAML/JSON for Kolla Ansible")
    parser.add_argument("-ki", "--kolla-inventory", metavar="INVENTORY",
                        help="specify inventory host path "
                             "(default=$%s/inventory or %s/inventory) for "
                             "Kolla Ansible" %
                             (CONFIG_PATH_ENV, DEFAULT_CONFIG_PATH))
    parser.add_argument("-kl", "--kolla-limit", metavar="SUBSET",
                        help="further limit selected hosts to an additional "
                             "pattern")
    parser.add_argument("-kp", "--kolla-playbook", metavar="PLAYBOOK",
                        help="path to Ansible playbook file")
    parser.add_argument("--kolla-skip-tags", metavar="TAGS",
                        help="only run plays and tasks whose tags do not "
                             "match these values in Kolla Ansible")
    parser.add_argument("-kt", "--kolla-tags", metavar="TAGS",
                        help="only run plays and tasks tagged with these "
                             "values in Kolla Ansible")
    parser.add_argument("--kolla-venv", metavar="VENV", default=default_venv,
                        help="path to virtualenv where Kolla Ansible is "
                             "installed (default=$%s or $PWD/%s)" %
                             (VENV_PATH_ENV, DEFAULT_VENV_PATH))


def _get_inventory_path(parsed_args, inventory_filename):
    """Return the path to the Kolla inventory."""
    if parsed_args.kolla_inventory:
        return parsed_args.kolla_inventory
    else:
        return os.path.join(parsed_args.kolla_config_path, "inventory",
                            inventory_filename)


def _validate_args(parsed_args, inventory_filename):
    """Validate Kayobe Ansible arguments."""
    vault.enforce_single_password_source(parsed_args)
    result = utils.is_readable_dir(parsed_args.kolla_config_path)
    if not result["result"]:
        LOG.error("Kolla configuration path %s is invalid: %s",
                  parsed_args.kolla_config_path, result["message"])
        sys.exit(1)

    inventory = _get_inventory_path(parsed_args, inventory_filename)
    result = utils.is_readable_dir(inventory)
    if not result["result"]:
        # NOTE(mgoddard): Previously the inventory was a file, now it is a
        # directory to allow us to support inventory host_vars. Support both
        # formats for now.
        result_f = utils.is_readable_file(inventory)
        if not result_f["result"]:
            LOG.error("Kolla inventory %s is invalid: %s",
                      inventory, result["message"])
            sys.exit(1)

    result = utils.is_readable_dir(parsed_args.kolla_venv)
    if not result["result"]:
        LOG.error("Kolla virtualenv %s is invalid: %s",
                  parsed_args.kolla_venv, result["message"])
        sys.exit(1)

    if parsed_args.kolla_playbook:
        result = utils.is_readable_file(parsed_args.kolla_playbook)
        if not result["result"]:
            LOG.error("Kolla Ansible playbook %s is invalid: %s",
                      parsed_args.kolla_playbook, result["message"])
            sys.exit(1)


def build_args(parsed_args, command, inventory_filename, extra_vars=None,
               tags=None, verbose_level=None, extra_args=None, limit=None):
    """Build arguments required for running Kolla Ansible."""
    venv_activate = os.path.join(parsed_args.kolla_venv, "bin", "activate")
    cmd = [".", venv_activate, "&&"]
    cmd += ["kolla-ansible", command]
    if verbose_level:
        cmd += ["-" + "v" * verbose_level]
    if parsed_args.kolla_playbook:
        cmd += ["--playbook", parsed_args.kolla_playbook]
    cmd += vault.build_args(parsed_args, "--key")
    inventory = _get_inventory_path(parsed_args, inventory_filename)
    cmd += ["--inventory", inventory]
    if parsed_args.kolla_config_path != DEFAULT_CONFIG_PATH:
        cmd += ["--configdir", parsed_args.kolla_config_path]
        cmd += ["--passwords",
                os.path.join(parsed_args.kolla_config_path, "passwords.yml")]
    if parsed_args.kolla_extra_vars:
        for extra_var in parsed_args.kolla_extra_vars:
            # Don't quote or escape variables passed via the kayobe -e CLI
            # argument, to match Ansible's behaviour.
            cmd += ["-e", extra_var]
    if extra_vars:
        for extra_var_name, extra_var_value in extra_vars.items():
            # Quote and escape variables originating within the python CLI.
            extra_var_value = utils.quote_and_escape(extra_var_value)
            cmd += ["-e", "%s=%s" % (extra_var_name, extra_var_value)]
    if parsed_args.kolla_limit or limit:
        limit_arg = utils.intersect_limits(parsed_args.kolla_limit, limit)
        cmd += ["--limit", utils.quote_and_escape(limit_arg)]
    if parsed_args.kolla_skip_tags:
        cmd += ["--skip-tags", parsed_args.kolla_skip_tags]
    if parsed_args.kolla_tags or tags:
        all_tags = [t for t in [parsed_args.kolla_tags, tags] if t]
        cmd += ["--tags", ",".join(all_tags)]
    if extra_args:
        cmd += extra_args
    return cmd


def _get_environment(parsed_args):
    """Return an environment dict for executing Kolla Ansible."""
    env = os.environ.copy()
    vault.update_environment(parsed_args, env)
    # If a custom Ansible configuration file exists, use it. Allow
    # etc/kayobe/kolla/ansible.cfg or etc/kayobe/ansible.cfg.
    ansible_cfg_path = os.path.join(parsed_args.config_path, "kolla",
                                    "ansible.cfg")
    if utils.is_readable_file(ansible_cfg_path)["result"]:
        env.setdefault("ANSIBLE_CONFIG", ansible_cfg_path)
    else:
        ansible_cfg_path = os.path.join(parsed_args.config_path, "ansible.cfg")
        if utils.is_readable_file(ansible_cfg_path)["result"]:
            env.setdefault("ANSIBLE_CONFIG", ansible_cfg_path)
    # kolla-ansible allows passing additional arguments to ansible-playbook via
    # EXTRA_OPTS.
    if parsed_args.check or parsed_args.diff:
        extra_opts = env.setdefault("EXTRA_OPTS", "")
        if parsed_args.check and "--check" not in extra_opts:
            env["EXTRA_OPTS"] += " --check"
        if parsed_args.diff and "--diff" not in extra_opts:
            env["EXTRA_OPTS"] += " --diff"
    return env


def run(parsed_args, command, inventory_filename, extra_vars=None,
        tags=None, quiet=False, verbose_level=None, extra_args=None,
        limit=None):
    """Run a Kolla Ansible command."""
    _validate_args(parsed_args, inventory_filename)
    cmd = build_args(parsed_args, command,
                     inventory_filename=inventory_filename,
                     extra_vars=extra_vars, tags=tags,
                     verbose_level=verbose_level,
                     extra_args=extra_args,
                     limit=limit)
    env = _get_environment(parsed_args)
    try:
        utils.run_command(" ".join(cmd), quiet=quiet, shell=True, env=env)
    except subprocess.CalledProcessError as e:
        LOG.error("kolla-ansible %s exited %d", command, e.returncode)
        sys.exit(e.returncode)


def run_seed(*args, **kwargs):
    """Run a Kolla Ansible command using the seed inventory."""
    return run(*args, inventory_filename="seed", **kwargs)


def run_overcloud(*args, **kwargs):
    """Run a Kolla Ansible command using the overcloud inventory."""
    return run(*args, inventory_filename="overcloud", **kwargs)



#!/usr/bin/env python3

# Usage: run this script and copy the output to etc/kayobe/kolla.yml

# See also: tools/kolla-feature-flags.sh

import os
import pathlib

import yaml

script_dir = pathlib.Path(__file__).parent.absolute()
path = os.path.join(script_dir, "../ansible/roles/kolla-ansible/vars/main.yml")

with open(path) as f:
    vars = yaml.safe_load(f)
    for key in vars["kolla_feature_flags"]:
        print("#kolla_enable_%s:" % key)



# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import subprocess
import unittest
from unittest import mock

import yaml

from kayobe import exception
from kayobe import utils


class TestCase(unittest.TestCase):

    maxDiff = None

    @mock.patch.object(utils, "run_command")
    def test_galaxy_role_install(self, mock_run):
        utils.galaxy_role_install("/path/to/role/file", "/path/to/roles")
        mock_run.assert_called_once_with(["ansible-galaxy", "role", "install",
                                          "--roles-path", "/path/to/roles",
                                          "--role-file", "/path/to/role/file"])

    @mock.patch.object(utils, "run_command")
    def test_galaxy_role_install_failure(self, mock_run):
        mock_run.side_effect = subprocess.CalledProcessError(1, "command")
        self.assertRaises(SystemExit,
                          utils.galaxy_role_install, "/path/to/role/file",
                          "/path/to/roles")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "read_yaml_file")
    def test_galaxy_collection_install(self, mock_read, mock_run):
        mock_read.return_value = {"collections": []}
        utils.galaxy_collection_install("/path/to/collection/file",
                                        "/path/to/collections")
        mock_run.assert_called_once_with(["ansible-galaxy", "collection",
                                          "install", "--collections-path",
                                          "/path/to/collections",
                                          "--requirements-file",
                                          "/path/to/collection/file"])

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "read_yaml_file")
    def test_galaxy_collection_install_failure(self, mock_read, mock_run):
        mock_read.return_value = {"collections": []}
        mock_run.side_effect = subprocess.CalledProcessError(1, "command")
        self.assertRaises(SystemExit,
                          utils.galaxy_collection_install,
                          "/path/to/collection/file", "/path/to/collections")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "read_file")
    def test_galaxy_collection_read_failure(self, mock_read, mock_run):
        mock_read.side_effect = IOError
        self.assertRaises(SystemExit,
                          utils.galaxy_collection_install,
                          "/path/to/collection/file", "/path/to/collections")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "read_yaml_file")
    def test_galaxy_collection_no_collections(self, mock_read, mock_run):
        mock_read.return_value = {"roles": []}
        utils.galaxy_collection_install("/path/to/collection/file",
                                        "/path/to/collections")
        mock_run.assert_called_once_with(["ansible-galaxy", "collection",
                                          "install", "--collections-path",
                                          "/path/to/collections",
                                          "--requirements-file",
                                          "/path/to/collection/file"])

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "read_yaml_file")
    def test_galaxy_collection_legacy_format(self, mock_read, mock_run):
        mock_read.return_value = []
        utils.galaxy_collection_install("/path/to/collection/file",
                                        "/path/to/collections")
        self.assertFalse(mock_run.called)

    @mock.patch.object(utils, "run_command")
    def test_galaxy_remove(self, mock_run):
        utils.galaxy_remove(["role1", "role2"], "/path/to/roles")
        mock_run.assert_called_once_with(["ansible-galaxy", "role", "remove",
                                          "--roles-path", "/path/to/roles",
                                          "role1", "role2"])

    @mock.patch.object(utils, "run_command")
    def test_galaxy_remove_failure(self, mock_run):
        mock_run.side_effect = subprocess.CalledProcessError(1, "command")
        self.assertRaises(SystemExit,
                          utils.galaxy_remove, ["role1", "role2"],
                          "/path/to/roles")

    @mock.patch.object(utils, "read_file")
    @mock.patch.object(yaml, "safe_load", wraps=yaml.safe_load)
    def test_read_yaml_file(self, mock_load, mock_read):
        config = """---
key1: value1
key2: value2
"""
        mock_read.return_value = config
        result = utils.read_yaml_file("/path/to/file")
        self.assertEqual(result, {"key1": "value1", "key2": "value2"})
        mock_read.assert_called_once_with("/path/to/file")
        mock_load.assert_called_once_with(config)

    @mock.patch.object(utils, "read_file")
    def test_read_yaml_file_open_failure(self, mock_read):
        mock_read.side_effect = IOError
        self.assertRaises(SystemExit, utils.read_yaml_file, "/path/to/file")

    @mock.patch.object(utils, "read_file")
    def test_read_yaml_file_not_yaml(self, mock_read):
        mock_read.return_value = "[1{!"
        self.assertRaises(SystemExit, utils.read_yaml_file, "/path/to/file")

    @mock.patch.object(subprocess, "check_call")
    def test_run_command(self, mock_call):
        output = utils.run_command(["command", "to", "run"])
        mock_call.assert_called_once_with(["command", "to", "run"])
        self.assertIsNone(output)

    @mock.patch("kayobe.utils.open")
    @mock.patch.object(subprocess, "check_call")
    def test_run_command_quiet(self, mock_call, mock_open):
        mock_devnull = mock_open.return_value.__enter__.return_value
        output = utils.run_command(["command", "to", "run"], quiet=True)
        mock_call.assert_called_once_with(["command", "to", "run"],
                                          stdout=mock_devnull,
                                          stderr=mock_devnull)
        self.assertIsNone(output)

    @mock.patch.object(subprocess, "check_output")
    def test_run_command_check_output(self, mock_output):
        mock_output.return_value = "command output"
        output = utils.run_command(["command", "to", "run"], check_output=True)
        mock_output.assert_called_once_with(["command", "to", "run"])
        self.assertEqual(output, "command output")

    @mock.patch.object(subprocess, "check_call")
    def test_run_command_failure(self, mock_call):
        mock_call.side_effect = subprocess.CalledProcessError(1, "command")
        self.assertRaises(subprocess.CalledProcessError, utils.run_command,
                          ["command", "to", "run"])

    def test_quote_and_escape_no_whitespace(self):
        self.assertEqual("'foo'", utils.quote_and_escape("foo"))

    def test_quote_and_escape_whitespace(self):
        self.assertEqual("'foo bar'", utils.quote_and_escape("foo bar"))

    def test_quote_and_escape_whitespace_with_quotes(self):
        self.assertEqual("'foo '\\''bar'\\'''",
                         utils.quote_and_escape("foo 'bar'"))

    def test_quote_and_escape_non_string(self):
        self.assertEqual(True, utils.quote_and_escape(True))

    def test_escape_jinja(self):
        value = "string to escape"
        expected = "{{'c3RyaW5nIHRvIGVzY2FwZQ==' | b64decode }}"
        self.assertEqual(expected, utils.escape_jinja(value))

    def test_detect_install_prefix(self):
        tmp_path = os.path.realpath('/tmp')
        path = "%s/test/local/lib/python3.6/dist-packages" % tmp_path
        expected = os.path.normpath("%s/test/local/" % tmp_path)
        result = utils._detect_install_prefix(path)
        self.assertEqual(expected, os.path.normpath(result))

    def test_detect_install_prefix_multiple_lib_components(self):
        tmp_path = os.path.realpath('/tmp')
        path = "%s/lib/test/local/lib/python3.6/dist-packages" % tmp_path
        expected = os.path.normpath("%s/lib/test/local/" % tmp_path)
        result = utils._detect_install_prefix(path)
        self.assertEqual(expected, os.path.normpath(result))

    def test_intersect_limits_no_arg_no_cli(self):
        result = utils.intersect_limits(None, None)
        self.assertEqual("", result)

    def test_intersect_limits_arg_no_cli(self):
        result = utils.intersect_limits("foo", None)
        self.assertEqual("foo", result)

    def test_intersect_limits_arg_no_cli_comma(self):
        result = utils.intersect_limits("foo,bar", None)
        self.assertEqual("foo,bar", result)

    def test_intersect_limits_arg_no_cli_colon(self):
        result = utils.intersect_limits("foo:bar", None)
        self.assertEqual("foo:bar", result)

    def test_intersect_limits_arg_no_cli_colon_and_comma(self):
        self.assertRaises(exception.Error,
                          utils.intersect_limits, "foo:bar,baz", None)

    def test_intersect_limits_no_arg_cli(self):
        result = utils.intersect_limits(None, "foo")
        self.assertEqual("foo", result)

    def test_intersect_limits_arg_and_cli(self):
        result = utils.intersect_limits("foo", "bar")
        self.assertEqual("foo:&bar", result)

    def test_intersect_limits_arg_and_cli_comma(self):
        result = utils.intersect_limits("foo,bar", "baz")
        self.assertEqual("foo,bar,&baz", result)

    def test_intersect_limits_arg_and_cli_colon(self):
        result = utils.intersect_limits("foo:bar", "baz")
        self.assertEqual("foo:bar:&baz", result)



# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import errno
import os
import os.path
import shutil
import subprocess
import tempfile
import unittest
from unittest import mock

from kayobe import ansible
from kayobe import exception
from kayobe import utils
from kayobe import vault


@mock.patch.dict(os.environ, clear=True)
class TestCase(unittest.TestCase):

    maxDiff = None

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    def test_run_playbooks(self, mock_validate, mock_vars, mock_run):
        mock_vars.return_value = ["/etc/kayobe/vars-file1.yml",
                                  "/etc/kayobe/vars-file2.yaml"]
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        ansible.run_playbooks(parsed_args, ["playbook1.yml", "playbook2.yml"])
        expected_cmd = [
            "ansible-playbook",
            "--inventory", utils.get_data_files_path("ansible", "inventory"),
            "--inventory", "/etc/kayobe/inventory",
            "-e", "@/etc/kayobe/vars-file1.yml",
            "-e", "@/etc/kayobe/vars-file2.yaml",
            "playbook1.yml",
            "playbook2.yml",
        ]
        home = os.path.expanduser("~")
        expected_env = {
            "KAYOBE_CONFIG_PATH": "/etc/kayobe",
            "ANSIBLE_ROLES_PATH": ":".join([
                "/etc/kayobe/ansible/roles",
                utils.get_data_files_path("ansible", "roles"),
                home + "/.ansible/roles",
                "/usr/share/ansible/roles",
                "/etc/ansible/roles",
            ]),
            "ANSIBLE_COLLECTIONS_PATH": ":".join([
                "/etc/kayobe/ansible/collections",
                utils.get_data_files_path("ansible", "collections"),
                home + "/.ansible/collections",
                "/usr/share/ansible/collections",
            ]),
            "ANSIBLE_ACTION_PLUGINS": ":".join([
                "/etc/kayobe/ansible/action_plugins",
                utils.get_data_files_path("ansible", "action_plugins"),
                home + "/.ansible/plugins/action",
                "/usr/share/ansible/plugins/action",
            ]),
            "ANSIBLE_FILTER_PLUGINS": ":".join([
                "/etc/kayobe/ansible/filter_plugins",
                utils.get_data_files_path("ansible", "filter_plugins"),
                home + "/.ansible/plugins/filter",
                "/usr/share/ansible/plugins/filter",
            ]),
            "ANSIBLE_TEST_PLUGINS": ":".join([
                "/etc/kayobe/ansible/test_plugins",
                utils.get_data_files_path("ansible", "test_plugins"),
                home + "/.ansible/plugins/test",
                "/usr/share/ansible/plugins/test",
            ]),
        }
        mock_run.assert_called_once_with(expected_cmd, check_output=False,
                                         quiet=False, env=expected_env)
        mock_vars.assert_called_once_with(["/etc/kayobe"])

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    def test_run_playbooks_all_the_args(self, mock_validate, mock_vars,
                                        mock_run):
        mock_vars.return_value = ["/path/to/config/vars-file1.yml",
                                  "/path/to/config/vars-file2.yaml"]
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        args = [
            "-b",
            "-C",
            "--config-path", "/path/to/config",
            "-D",
            "--environment", "test-env",
            "-e", "ev_name1=ev_value1",
            "-i", "/path/to/inventory",
            "-l", "group1:host",
            "-t", "tag1,tag2",
            "-lt",
        ]
        parsed_args = parser.parse_args(args)
        ansible.run_playbooks(parsed_args, ["playbook1.yml", "playbook2.yml"],
                              verbose_level=2)
        expected_cmd = [
            "ansible-playbook",
            "-vv",
            "--list-tasks",
            "--inventory", utils.get_data_files_path("ansible", "inventory"),
            "--inventory", "/path/to/inventory",
            "-e", "@/path/to/config/vars-file1.yml",
            "-e", "@/path/to/config/vars-file2.yaml",
            "-e", "ev_name1=ev_value1",
            "--become",
            "--check",
            "--diff",
            "--limit", "group1:host",
            "--tags", "tag1,tag2",
            "playbook1.yml",
            "playbook2.yml",
        ]
        home = os.path.expanduser("~")
        expected_env = {
            "KAYOBE_CONFIG_PATH": "/path/to/config",
            "KAYOBE_ENVIRONMENT": "test-env",
            "ANSIBLE_ROLES_PATH": ":".join([
                "/path/to/config/ansible/roles",
                utils.get_data_files_path("ansible", "roles"),
                home + "/.ansible/roles",
                "/usr/share/ansible/roles",
                "/etc/ansible/roles",
            ]),
            "ANSIBLE_COLLECTIONS_PATH": ":".join([
                "/path/to/config/ansible/collections",
                utils.get_data_files_path("ansible", "collections"),
                home + "/.ansible/collections",
                "/usr/share/ansible/collections",
            ]),
            "ANSIBLE_ACTION_PLUGINS": ":".join([
                "/path/to/config/ansible/action_plugins",
                utils.get_data_files_path("ansible", "action_plugins"),
                home + "/.ansible/plugins/action",
                "/usr/share/ansible/plugins/action",
            ]),
            "ANSIBLE_FILTER_PLUGINS": ":".join([
                "/path/to/config/ansible/filter_plugins",
                utils.get_data_files_path("ansible", "filter_plugins"),
                home + "/.ansible/plugins/filter",
                "/usr/share/ansible/plugins/filter",
            ]),
            "ANSIBLE_TEST_PLUGINS": ":".join([
                "/path/to/config/ansible/test_plugins",
                utils.get_data_files_path("ansible", "test_plugins"),
                home + "/.ansible/plugins/test",
                "/usr/share/ansible/plugins/test",
            ]),
        }
        mock_run.assert_called_once_with(expected_cmd, check_output=False,
                                         quiet=False, env=expected_env)
        mock_vars.assert_called_once_with(
            ["/path/to/config", "/path/to/config/environments/test-env"])

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    @mock.patch.object(vault, "_ask_vault_pass")
    def test_run_playbooks_all_the_long_args(self, mock_ask, mock_validate,
                                             mock_vars, mock_run):
        mock_vars.return_value = ["/path/to/config/vars-file1.yml",
                                  "/path/to/config/vars-file2.yaml"]
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        mock_ask.return_value = "test-pass"
        args = [
            "--ask-vault-pass",
            "--become",
            "--check",
            "--config-path", "/path/to/config",
            "--diff",
            "--environment", "test-env",
            "--extra-vars", "ev_name1=ev_value1",
            "--inventory", "/path/to/inventory",
            "--limit", "group1:host1",
            "--skip-tags", "tag3,tag4",
            "--tags", "tag1,tag2",
            "--list-tasks",
        ]
        parsed_args = parser.parse_args(args)
        mock_run.return_value = "/path/to/kayobe-vault-password-helper"
        ansible.run_playbooks(parsed_args, ["playbook1.yml", "playbook2.yml"])
        expected_cmd = [
            "ansible-playbook",
            "--list-tasks",
            "--vault-password-file", "/path/to/kayobe-vault-password-helper",
            "--inventory", utils.get_data_files_path("ansible", "inventory"),
            "--inventory", "/path/to/inventory",
            "-e", "@/path/to/config/vars-file1.yml",
            "-e", "@/path/to/config/vars-file2.yaml",
            "-e", "ev_name1=ev_value1",
            "--become",
            "--check",
            "--diff",
            "--limit", "group1:host1",
            "--skip-tags", "tag3,tag4",
            "--tags", "tag1,tag2",
            "playbook1.yml",
            "playbook2.yml",
        ]
        expected_env = {
            "KAYOBE_CONFIG_PATH": "/path/to/config",
            "KAYOBE_ENVIRONMENT": "test-env",
            "KAYOBE_VAULT_PASSWORD": "test-pass",
            "ANSIBLE_ROLES_PATH": mock.ANY,
            "ANSIBLE_COLLECTIONS_PATH": mock.ANY,
            "ANSIBLE_ACTION_PLUGINS": mock.ANY,
            "ANSIBLE_FILTER_PLUGINS": mock.ANY,
            "ANSIBLE_TEST_PLUGINS": mock.ANY,
        }
        expected_calls = [
            mock.call(["which", "kayobe-vault-password-helper"],
                      check_output=True, universal_newlines=True),
            mock.call(expected_cmd, check_output=False, quiet=False,
                      env=expected_env)
        ]
        self.assertListEqual(expected_calls, mock_run.mock_calls)
        mock_vars.assert_called_once_with(
            ["/path/to/config", "/path/to/config/environments/test-env"])

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    @mock.patch.object(vault, "update_environment")
    def test_run_playbooks_vault_password_file(self, mock_update,
                                               mock_validate,
                                               mock_vars, mock_run):
        mock_vars.return_value = []
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        args = [
            "--vault-password-file", "/path/to/vault/pw",
        ]
        parsed_args = parser.parse_args(args)
        ansible.run_playbooks(parsed_args, ["playbook1.yml"])
        expected_cmd = [
            "ansible-playbook",
            "--vault-password-file", "/path/to/vault/pw",
            "--inventory", utils.get_data_files_path("ansible", "inventory"),
            "--inventory", "/etc/kayobe/inventory",
            "playbook1.yml",
        ]
        expected_env = {
            "KAYOBE_CONFIG_PATH": "/etc/kayobe",
            "ANSIBLE_ROLES_PATH": mock.ANY,
            "ANSIBLE_COLLECTIONS_PATH": mock.ANY,
            "ANSIBLE_ACTION_PLUGINS": mock.ANY,
            "ANSIBLE_FILTER_PLUGINS": mock.ANY,
            "ANSIBLE_TEST_PLUGINS": mock.ANY,
        }
        mock_run.assert_called_once_with(expected_cmd, check_output=False,
                                         quiet=False, env=expected_env)
        mock_update.assert_called_once_with(mock.ANY, expected_env)

    @mock.patch.dict(os.environ, {"KAYOBE_VAULT_PASSWORD": "test-pass"},
                     clear=True)
    @mock.patch.object(utils, "run_command")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    def test_run_playbooks_vault_password_helper(self, mock_validate,
                                                 mock_vars, mock_run):
        mock_vars.return_value = []
        parser = argparse.ArgumentParser()
        mock_run.return_value = "/path/to/kayobe-vault-password-helper"
        ansible.add_args(parser)
        vault.add_args(parser)
        mock_run.assert_called_once_with(
            ["which", "kayobe-vault-password-helper"], check_output=True,
            universal_newlines=True)
        mock_run.reset_mock()
        parsed_args = parser.parse_args([])
        ansible.run_playbooks(parsed_args, ["playbook1.yml"])
        expected_cmd = [
            "ansible-playbook",
            "--vault-password-file", "/path/to/kayobe-vault-password-helper",
            "--inventory", utils.get_data_files_path("ansible", "inventory"),
            "--inventory", "/etc/kayobe/inventory",
            "playbook1.yml",
        ]
        expected_env = {
            "KAYOBE_CONFIG_PATH": "/etc/kayobe",
            "KAYOBE_VAULT_PASSWORD": "test-pass",
            "ANSIBLE_ROLES_PATH": mock.ANY,
            "ANSIBLE_COLLECTIONS_PATH": mock.ANY,
            "ANSIBLE_ACTION_PLUGINS": mock.ANY,
            "ANSIBLE_FILTER_PLUGINS": mock.ANY,
            "ANSIBLE_TEST_PLUGINS": mock.ANY,
        }
        mock_run.assert_called_once_with(expected_cmd, check_output=False,
                                         quiet=False, env=expected_env)

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    def test_run_playbooks_vault_ask_and_file(self, mock_validate, mock_vars,
                                              mock_run):
        mock_vars.return_value = []
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        args = [
            "--ask-vault-pass",
            "--vault-password-file", "/path/to/vault/pw",
        ]
        self.assertRaises(SystemExit, parser.parse_args, args)

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    def test_run_playbooks_func_args(self, mock_validate, mock_vars, mock_run):
        mock_vars.return_value = ["/etc/kayobe/vars-file1.yml",
                                  "/etc/kayobe/vars-file2.yaml"]
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        args = [
            "--extra-vars", "ev_name1=ev_value1",
            "--limit", "group1:host1",
            "--tags", "tag1,tag2",
        ]
        parsed_args = parser.parse_args(args)
        kwargs = {
            "extra_vars": {"ev_name2": "ev_value2"},
            "limit": "group2:host2",
            "tags": "tag3,tag4",
            "verbose_level": 0,
            "check": True,
            "diff": True,
        }
        ansible.run_playbooks(parsed_args, ["playbook1.yml", "playbook2.yml"],
                              **kwargs)
        expected_cmd = [
            "ansible-playbook",
            "--inventory", utils.get_data_files_path("ansible", "inventory"),
            "--inventory", "/etc/kayobe/inventory",
            "-e", "@/etc/kayobe/vars-file1.yml",
            "-e", "@/etc/kayobe/vars-file2.yaml",
            "-e", "ev_name1=ev_value1",
            "-e", "ev_name2='ev_value2'",
            "--check",
            "--diff",
            "--limit", "group1:host1:&group2:host2",
            "--tags", "tag1,tag2,tag3,tag4",
            "playbook1.yml",
            "playbook2.yml",
        ]
        expected_env = {
            "KAYOBE_CONFIG_PATH": "/etc/kayobe",
            "ANSIBLE_ROLES_PATH": mock.ANY,
            "ANSIBLE_COLLECTIONS_PATH": mock.ANY,
            "ANSIBLE_ACTION_PLUGINS": mock.ANY,
            "ANSIBLE_FILTER_PLUGINS": mock.ANY,
            "ANSIBLE_TEST_PLUGINS": mock.ANY,
        }
        mock_run.assert_called_once_with(expected_cmd, check_output=False,
                                         quiet=False, env=expected_env)
        mock_vars.assert_called_once_with(["/etc/kayobe"])

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    def test_run_playbooks_ignore_limit(self, mock_validate, mock_vars,
                                        mock_run):
        mock_vars.return_value = ["/etc/kayobe/vars-file1.yml",
                                  "/etc/kayobe/vars-file2.yaml"]
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        args = [
            "-l", "group1:host",
        ]
        parsed_args = parser.parse_args(args)
        ansible.run_playbooks(parsed_args, ["playbook1.yml", "playbook2.yml"],
                              limit="foo", ignore_limit=True)
        expected_cmd = [
            "ansible-playbook",
            "--inventory", utils.get_data_files_path("ansible", "inventory"),
            "--inventory", "/etc/kayobe/inventory",
            "-e", "@/etc/kayobe/vars-file1.yml",
            "-e", "@/etc/kayobe/vars-file2.yaml",
            "playbook1.yml",
            "playbook2.yml",
        ]
        expected_env = {
            "KAYOBE_CONFIG_PATH": "/etc/kayobe",
            "ANSIBLE_ROLES_PATH": mock.ANY,
            "ANSIBLE_COLLECTIONS_PATH": mock.ANY,
            "ANSIBLE_ACTION_PLUGINS": mock.ANY,
            "ANSIBLE_FILTER_PLUGINS": mock.ANY,
            "ANSIBLE_TEST_PLUGINS": mock.ANY,
        }
        mock_run.assert_called_once_with(expected_cmd, check_output=False,
                                         quiet=False, env=expected_env)
        mock_vars.assert_called_once_with(["/etc/kayobe"])

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    def test_run_playbooks_list_tasks_arg(self, mock_validate, mock_vars,
                                          mock_run):
        mock_vars.return_value = []
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        args = [
            "--list-tasks",
        ]
        parsed_args = parser.parse_args(args)
        kwargs = {
            "list_tasks": False
        }
        ansible.run_playbooks(parsed_args, ["playbook1.yml", "playbook2.yml"],
                              **kwargs)
        expected_cmd = [
            "ansible-playbook",
            "--inventory", utils.get_data_files_path("ansible", "inventory"),
            "--inventory", "/etc/kayobe/inventory",
            "playbook1.yml",
            "playbook2.yml",
        ]
        expected_env = {
            "KAYOBE_CONFIG_PATH": "/etc/kayobe",
            "ANSIBLE_ROLES_PATH": mock.ANY,
            "ANSIBLE_COLLECTIONS_PATH": mock.ANY,
            "ANSIBLE_ACTION_PLUGINS": mock.ANY,
            "ANSIBLE_FILTER_PLUGINS": mock.ANY,
            "ANSIBLE_TEST_PLUGINS": mock.ANY,
        }
        mock_run.assert_called_once_with(expected_cmd, check_output=False,
                                         quiet=False, env=expected_env)
        mock_vars.assert_called_once_with(["/etc/kayobe"])

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "is_readable_file")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    def test_run_playbooks_ansible_cfg(self, mock_validate, mock_vars,
                                       mock_readable, mock_run):
        mock_vars.return_value = []
        mock_readable.return_value = {"result": True}
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        ansible.run_playbooks(parsed_args, ["playbook1.yml"])
        expected_cmd = [
            "ansible-playbook",
            "--inventory", utils.get_data_files_path("ansible", "inventory"),
            "--inventory", "/etc/kayobe/inventory",
            "playbook1.yml",
        ]
        expected_env = {
            "ANSIBLE_CONFIG": "/etc/kayobe/ansible.cfg",
            "KAYOBE_CONFIG_PATH": "/etc/kayobe",
            "ANSIBLE_ROLES_PATH": mock.ANY,
            "ANSIBLE_COLLECTIONS_PATH": mock.ANY,
            "ANSIBLE_ACTION_PLUGINS": mock.ANY,
            "ANSIBLE_FILTER_PLUGINS": mock.ANY,
            "ANSIBLE_TEST_PLUGINS": mock.ANY,
        }
        mock_run.assert_called_once_with(expected_cmd, check_output=False,
                                         quiet=False, env=expected_env)
        mock_vars.assert_called_once_with(["/etc/kayobe"])
        mock_readable.assert_called_once_with("/etc/kayobe/ansible.cfg")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "is_readable_file")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    def test_run_playbooks_ansible_cfg_env(self, mock_validate, mock_vars,
                                           mock_readable, mock_run):
        mock_vars.return_value = []
        mock_readable.return_value = {"result": True}
        os.environ["ANSIBLE_CONFIG"] = "/path/to/ansible.cfg"
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        ansible.run_playbooks(parsed_args, ["playbook1.yml"])
        expected_cmd = [
            "ansible-playbook",
            "--inventory", utils.get_data_files_path("ansible", "inventory"),
            "--inventory", "/etc/kayobe/inventory",
            "playbook1.yml",
        ]
        expected_env = {
            "ANSIBLE_CONFIG": "/path/to/ansible.cfg",
            "KAYOBE_CONFIG_PATH": "/etc/kayobe",
            "ANSIBLE_ROLES_PATH": mock.ANY,
            "ANSIBLE_COLLECTIONS_PATH": mock.ANY,
            "ANSIBLE_ACTION_PLUGINS": mock.ANY,
            "ANSIBLE_FILTER_PLUGINS": mock.ANY,
            "ANSIBLE_TEST_PLUGINS": mock.ANY,
        }
        mock_run.assert_called_once_with(expected_cmd, check_output=False,
                                         quiet=False, env=expected_env)
        mock_vars.assert_called_once_with(["/etc/kayobe"])
        mock_readable.assert_called_once_with("/etc/kayobe/ansible.cfg")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    def test_run_playbooks_failure(self, mock_validate, mock_vars, mock_run):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        mock_run.side_effect = subprocess.CalledProcessError(1, "dummy")
        self.assertRaises(SystemExit,
                          ansible.run_playbooks, parsed_args, ["command"])

    @mock.patch.object(shutil, 'rmtree')
    @mock.patch.object(utils, 'read_yaml_file')
    @mock.patch.object(os, 'listdir')
    @mock.patch.object(ansible, 'run_playbook')
    @mock.patch.object(tempfile, 'mkdtemp')
    def test_config_dump(self, mock_mkdtemp, mock_run, mock_listdir, mock_read,
                         mock_rmtree):
        parser = argparse.ArgumentParser()
        parsed_args = parser.parse_args([])
        dump_dir = "/path/to/dump"
        mock_mkdtemp.return_value = dump_dir
        mock_listdir.return_value = ["host1.yml", "host2.yml"]
        mock_read.side_effect = [
            {"var1": "value1"},
            {"var2": "value2"}
        ]
        result = ansible.config_dump(parsed_args)
        expected_result = {
            "host1": {"var1": "value1"},
            "host2": {"var2": "value2"},
        }
        self.assertEqual(result, expected_result)
        dump_config_path = utils.get_data_files_path(
            "ansible", "dump-config.yml")
        mock_run.assert_called_once_with(parsed_args,
                                         dump_config_path,
                                         extra_vars={
                                             "dump_path": dump_dir,
                                         },
                                         check_output=True, tags=None,
                                         verbose_level=None, check=False,
                                         list_tasks=False, diff=False)
        mock_rmtree.assert_called_once_with(dump_dir)
        mock_listdir.assert_any_call(dump_dir)
        mock_read.assert_has_calls([
            mock.call(os.path.join(dump_dir, "host1.yml")),
            mock.call(os.path.join(dump_dir, "host2.yml")),
        ])

    @mock.patch.object(utils, 'galaxy_role_install', autospec=True)
    @mock.patch.object(utils, 'is_readable_file', autospec=True)
    @mock.patch.object(os, 'makedirs', autospec=True)
    def test_install_galaxy_roles(self, mock_mkdirs, mock_is_readable,
                                  mock_install):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        parsed_args = parser.parse_args([])
        mock_is_readable.return_value = {"result": False}

        ansible.install_galaxy_roles(parsed_args)

        mock_install.assert_called_once_with(utils.get_data_files_path(
            "requirements.yml"), utils.get_data_files_path(
            "ansible", "roles"), force=False)
        mock_is_readable.assert_called_once_with(
            "/etc/kayobe/ansible/requirements.yml")
        self.assertFalse(mock_mkdirs.called)

    @mock.patch.object(utils, 'galaxy_role_install', autospec=True)
    @mock.patch.object(utils, 'is_readable_file', autospec=True)
    @mock.patch.object(os, 'makedirs', autospec=True)
    def test_install_galaxy_roles_with_kayobe_config(
            self, mock_mkdirs, mock_is_readable, mock_install):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        parsed_args = parser.parse_args([])
        mock_is_readable.return_value = {"result": True}

        ansible.install_galaxy_roles(parsed_args)

        expected_calls = [
            mock.call(utils.get_data_files_path("requirements.yml"),
                      utils.get_data_files_path("ansible", "roles"),
                      force=False),
            mock.call("/etc/kayobe/ansible/requirements.yml",
                      "/etc/kayobe/ansible/roles", force=False)]
        self.assertListEqual(expected_calls, mock_install.call_args_list)
        mock_is_readable.assert_called_once_with(
            "/etc/kayobe/ansible/requirements.yml")
        mock_mkdirs.assert_called_once_with("/etc/kayobe/ansible/roles")

    @mock.patch.object(utils, 'galaxy_role_install', autospec=True)
    @mock.patch.object(utils, 'is_readable_file', autospec=True)
    @mock.patch.object(os, 'makedirs', autospec=True)
    def test_install_galaxy_roles_with_kayobe_config_forced(
            self, mock_mkdirs, mock_is_readable, mock_install):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        parsed_args = parser.parse_args([])
        mock_is_readable.return_value = {"result": True}

        ansible.install_galaxy_roles(parsed_args, force=True)

        expected_calls = [
            mock.call(utils.get_data_files_path("requirements.yml"),
                      utils.get_data_files_path("ansible", "roles"),
                      force=True),
            mock.call("/etc/kayobe/ansible/requirements.yml",
                      "/etc/kayobe/ansible/roles", force=True)]
        self.assertListEqual(expected_calls, mock_install.call_args_list)
        mock_is_readable.assert_called_once_with(
            "/etc/kayobe/ansible/requirements.yml")
        mock_mkdirs.assert_called_once_with("/etc/kayobe/ansible/roles")

    @mock.patch.object(utils, 'galaxy_role_install', autospec=True)
    @mock.patch.object(utils, 'is_readable_file', autospec=True)
    @mock.patch.object(os, 'makedirs', autospec=True)
    def test_install_galaxy_roles_with_kayobe_config_mkdirs_failure(
            self, mock_mkdirs, mock_is_readable, mock_install):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        parsed_args = parser.parse_args([])
        mock_is_readable.return_value = {"result": True}
        mock_mkdirs.side_effect = OSError(errno.EPERM)

        self.assertRaises(exception.Error,
                          ansible.install_galaxy_roles, parsed_args)

        mock_install.assert_called_once_with(utils.get_data_files_path(
            "requirements.yml"), utils.get_data_files_path("ansible", "roles"),
            force=False)
        mock_is_readable.assert_called_once_with(
            "/etc/kayobe/ansible/requirements.yml")
        mock_mkdirs.assert_called_once_with("/etc/kayobe/ansible/roles")

    @mock.patch.object(utils, 'galaxy_collection_install', autospec=True)
    @mock.patch.object(utils, 'is_readable_file', autospec=True)
    @mock.patch.object(os, 'makedirs', autospec=True)
    def test_install_galaxy_collections(self, mock_mkdirs, mock_is_readable,
                                        mock_install):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        parsed_args = parser.parse_args([])
        mock_is_readable.return_value = {"result": False}

        ansible.install_galaxy_collections(parsed_args)

        mock_install.assert_called_once_with(utils.get_data_files_path(
            "requirements.yml"), utils.get_data_files_path(
            "ansible", "collections"), force=False)
        mock_is_readable.assert_called_once_with(
            "/etc/kayobe/ansible/requirements.yml")
        self.assertFalse(mock_mkdirs.called)

    @mock.patch.object(utils, 'galaxy_collection_install', autospec=True)
    @mock.patch.object(utils, 'is_readable_file', autospec=True)
    @mock.patch.object(os, 'makedirs', autospec=True)
    def test_install_galaxy_collections_with_kayobe_config(
            self, mock_mkdirs, mock_is_readable, mock_install):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        parsed_args = parser.parse_args([])
        mock_is_readable.return_value = {"result": True}

        ansible.install_galaxy_collections(parsed_args)

        expected_calls = [
            mock.call(utils.get_data_files_path("requirements.yml"),
                      utils.get_data_files_path("ansible", "collections"),
                      force=False),
            mock.call("/etc/kayobe/ansible/requirements.yml",
                      "/etc/kayobe/ansible/collections", force=False)]
        self.assertListEqual(expected_calls, mock_install.call_args_list)
        mock_is_readable.assert_called_once_with(
            "/etc/kayobe/ansible/requirements.yml")
        mock_mkdirs.assert_called_once_with("/etc/kayobe/ansible/collections")

    @mock.patch.object(utils, 'galaxy_collection_install', autospec=True)
    @mock.patch.object(utils, 'is_readable_file', autospec=True)
    @mock.patch.object(os, 'makedirs', autospec=True)
    def test_install_galaxy_collections_with_kayobe_config_forced(
            self, mock_mkdirs, mock_is_readable, mock_install):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        parsed_args = parser.parse_args([])
        mock_is_readable.return_value = {"result": True}

        ansible.install_galaxy_collections(parsed_args, force=True)

        expected_calls = [
            mock.call(utils.get_data_files_path("requirements.yml"),
                      utils.get_data_files_path("ansible", "collections"),
                      force=True),
            mock.call("/etc/kayobe/ansible/requirements.yml",
                      "/etc/kayobe/ansible/collections", force=True)]
        self.assertListEqual(expected_calls, mock_install.call_args_list)
        mock_is_readable.assert_called_once_with(
            "/etc/kayobe/ansible/requirements.yml")
        mock_mkdirs.assert_called_once_with("/etc/kayobe/ansible/collections")

    @mock.patch.object(utils, 'galaxy_collection_install', autospec=True)
    @mock.patch.object(utils, 'is_readable_file', autospec=True)
    @mock.patch.object(os, 'makedirs', autospec=True)
    def test_install_galaxy_collections_with_kayobe_config_mkdirs_failure(
            self, mock_mkdirs, mock_is_readable, mock_install):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        parsed_args = parser.parse_args([])
        mock_is_readable.return_value = {"result": True}
        mock_mkdirs.side_effect = OSError(errno.EPERM)

        self.assertRaises(exception.Error,
                          ansible.install_galaxy_collections, parsed_args)

        mock_install.assert_called_once_with(
            utils.get_data_files_path("requirements.yml"),
            utils.get_data_files_path("ansible", "collections"), force=False)
        mock_is_readable.assert_called_once_with(
            "/etc/kayobe/ansible/requirements.yml")
        mock_mkdirs.assert_called_once_with("/etc/kayobe/ansible/collections")

    @mock.patch.object(utils, 'galaxy_remove', autospec=True)
    def test_prune_galaxy_roles(self, mock_remove):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        parsed_args = parser.parse_args([])

        ansible.prune_galaxy_roles(parsed_args)

        expected_roles = [
            'resmo.ntp',
            'stackhpc.ntp',
            'stackhpc.os-shade',
            'yatesr.timezone',
        ]
        mock_remove.assert_called_once_with(expected_roles,
                                            "ansible/roles")

    @mock.patch.object(utils, 'is_readable_file', autospec=True)
    def test_passwords_yml_exists_false(self, mock_is_readable):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        parsed_args = parser.parse_args([])
        mock_is_readable.return_value = {"result": False}

        result = ansible.passwords_yml_exists(parsed_args)

        self.assertFalse(result)
        mock_is_readable.assert_called_once_with(
            "/etc/kayobe/kolla/passwords.yml")

    @mock.patch.object(utils, 'is_readable_file', autospec=True)
    def test_passwords_yml_exists_true(self, mock_is_readable):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        parsed_args = parser.parse_args(["--config-path", "/path/to/config"])
        mock_is_readable.return_value = {"result": True}

        result = ansible.passwords_yml_exists(parsed_args)

        self.assertTrue(result)
        mock_is_readable.assert_called_once_with(
            "/path/to/config/kolla/passwords.yml")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    def test_multiple_inventory_args(self, mock_validate, mock_vars, mock_run):
        mock_vars.return_value = []
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        args = [
            "--inventory", "/etc/kayobe/inventory",
            "--inventory", "/etc/kayobe/environments/foobar/inventory",
        ]
        parsed_args = parser.parse_args(args)
        ansible.run_playbooks(parsed_args, ["playbook1.yml", "playbook2.yml"])
        expected_cmd = [
            "ansible-playbook",
            "--inventory", utils.get_data_files_path("ansible", "inventory"),
            "--inventory", "/etc/kayobe/inventory",
            "--inventory", "/etc/kayobe/environments/foobar/inventory",
            "playbook1.yml",
            "playbook2.yml",
        ]
        expected_env = {
            "KAYOBE_CONFIG_PATH": "/etc/kayobe",
            "ANSIBLE_ROLES_PATH": mock.ANY,
            "ANSIBLE_COLLECTIONS_PATH": mock.ANY,
            "ANSIBLE_ACTION_PLUGINS": mock.ANY,
            "ANSIBLE_FILTER_PLUGINS": mock.ANY,
            "ANSIBLE_TEST_PLUGINS": mock.ANY,
        }
        mock_run.assert_called_once_with(expected_cmd, check_output=False,
                                         quiet=False, env=expected_env)
        mock_vars.assert_called_once_with(["/etc/kayobe"])

    @mock.patch.object(os.path, "exists")
    @mock.patch.object(utils, "run_command")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    def test_multiple_inventories(self, mock_validate, mock_vars, mock_run,
                                  mock_exists):
        mock_vars.return_value = []
        # os.path.exists gets called three times:
        # 1) shared inventory
        # 2) environment inventory
        # 3) ansible.cfg
        mock_exists.side_effect = [True, True, False]
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        args = [
            "--environment", "test-env",
        ]
        parsed_args = parser.parse_args(args)
        ansible.run_playbooks(parsed_args, ["playbook1.yml", "playbook2.yml"])
        expected_cmd = [
            "ansible-playbook",
            "--inventory", utils.get_data_files_path("ansible", "inventory"),
            "--inventory", "/etc/kayobe/inventory",
            "--inventory", "/etc/kayobe/environments/test-env/inventory",
            "playbook1.yml",
            "playbook2.yml",
        ]
        expected_env = {
            "KAYOBE_CONFIG_PATH": "/etc/kayobe",
            "KAYOBE_ENVIRONMENT": "test-env",
            "ANSIBLE_ROLES_PATH": mock.ANY,
            "ANSIBLE_COLLECTIONS_PATH": mock.ANY,
            "ANSIBLE_ACTION_PLUGINS": mock.ANY,
            "ANSIBLE_FILTER_PLUGINS": mock.ANY,
            "ANSIBLE_TEST_PLUGINS": mock.ANY,
        }
        expected_calls = [
            mock.call("/etc/kayobe/inventory"),
            mock.call("/etc/kayobe/environments/test-env/inventory"),
            mock.call("/etc/kayobe/ansible.cfg"),
        ]
        self.assertListEqual(expected_calls, mock_exists.mock_calls)
        mock_run.assert_called_once_with(expected_cmd, check_output=False,
                                         quiet=False, env=expected_env)
        mock_vars.assert_called_once_with(
            ["/etc/kayobe", "/etc/kayobe/environments/test-env"])

    @mock.patch.object(os.path, "exists")
    @mock.patch.object(utils, "run_command")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    def test_shared_inventory_only(self, mock_validate, mock_vars, mock_run,
                                   mock_exists):
        mock_vars.return_value = []
        # os.path.exists gets called three times:
        # 1) shared inventory
        # 2) environment inventory
        # 3) ansible.cfg
        mock_exists.side_effect = [True, False, False]
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        args = [
            "--environment", "test-env",
        ]
        parsed_args = parser.parse_args(args)
        ansible.run_playbooks(parsed_args, ["playbook1.yml", "playbook2.yml"])
        expected_cmd = [
            "ansible-playbook",
            "--inventory", utils.get_data_files_path("ansible", "inventory"),
            "--inventory", "/etc/kayobe/inventory",
            "playbook1.yml",
            "playbook2.yml",
        ]
        expected_env = {
            "KAYOBE_CONFIG_PATH": "/etc/kayobe",
            "KAYOBE_ENVIRONMENT": "test-env",
            "ANSIBLE_ROLES_PATH": mock.ANY,
            "ANSIBLE_COLLECTIONS_PATH": mock.ANY,
            "ANSIBLE_ACTION_PLUGINS": mock.ANY,
            "ANSIBLE_FILTER_PLUGINS": mock.ANY,
            "ANSIBLE_TEST_PLUGINS": mock.ANY,
        }
        expected_calls = [
            mock.call("/etc/kayobe/inventory"),
            mock.call("/etc/kayobe/environments/test-env/inventory"),
            mock.call("/etc/kayobe/ansible.cfg"),
        ]
        self.assertListEqual(expected_calls, mock_exists.mock_calls)
        mock_run.assert_called_once_with(expected_cmd, check_output=False,
                                         quiet=False, env=expected_env)
        mock_vars.assert_called_once_with(
            ["/etc/kayobe", "/etc/kayobe/environments/test-env"])

    @mock.patch.object(os.path, "exists")
    @mock.patch.object(utils, "run_command")
    @mock.patch.object(ansible, "_get_vars_files")
    @mock.patch.object(ansible, "_validate_args")
    def test_env_inventory_only(self, mock_validate, mock_vars, mock_run,
                                mock_exists):
        mock_vars.return_value = []
        # os.path.exists gets called three times:
        # 1) shared inventory
        # 2) environment inventory
        # 3) ansible.cfg
        mock_exists.side_effect = [False, True, False]
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        vault.add_args(parser)
        args = [
            "--environment", "test-env",
        ]
        parsed_args = parser.parse_args(args)
        ansible.run_playbooks(parsed_args, ["playbook1.yml", "playbook2.yml"])
        expected_cmd = [
            "ansible-playbook",
            "--inventory", utils.get_data_files_path("ansible", "inventory"),
            "--inventory", "/etc/kayobe/environments/test-env/inventory",
            "playbook1.yml",
            "playbook2.yml",
        ]
        expected_env = {
            "KAYOBE_CONFIG_PATH": "/etc/kayobe",
            "KAYOBE_ENVIRONMENT": "test-env",
            "ANSIBLE_ROLES_PATH": mock.ANY,
            "ANSIBLE_COLLECTIONS_PATH": mock.ANY,
            "ANSIBLE_ACTION_PLUGINS": mock.ANY,
            "ANSIBLE_FILTER_PLUGINS": mock.ANY,
            "ANSIBLE_TEST_PLUGINS": mock.ANY,
        }
        expected_calls = [
            mock.call("/etc/kayobe/inventory"),
            mock.call("/etc/kayobe/environments/test-env/inventory"),
            mock.call("/etc/kayobe/ansible.cfg"),
        ]
        self.assertListEqual(expected_calls, mock_exists.mock_calls)
        mock_run.assert_called_once_with(expected_cmd, check_output=False,
                                         quiet=False, env=expected_env)
        mock_vars.assert_called_once_with(
            ["/etc/kayobe", "/etc/kayobe/environments/test-env"])



# Copyright (c) 2021 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import os.path
import sys

from kayobe import utils


LOG = logging.getLogger(__name__)


def add_args(parser):
    """Add arguments required for managing Kayobe environments to a parser."""
    parser.add_argument("--source-config-path",
                        help="Kayobe configuration to import")


def create_kayobe_environment(parsed_args):
    """Create a new Kayobe environment."""
    if not parsed_args.environment:
        LOG.error("You must specify an environment to create")
        sys.exit(1)

    # Ensure environments directory exists and is readable inside config path
    kc_environments = os.path.join(parsed_args.config_path, "environments")
    result = utils.is_readable_dir(kc_environments)
    if not result["result"]:
        if result["message"] == "Path does not exist":
            os.mkdir(kc_environments)
        else:
            LOG.error("Kayobe global environments directory %s is invalid: %s",
                      kc_environments, result["message"])
            sys.exit(1)

    env_path = os.path.join(kc_environments, parsed_args.environment)
    result = utils.is_readable_dir(env_path)
    if result["result"]:
        LOG.error("Kayobe environment directory %s already exists", env_path)
        sys.exit(1)
    else:
        if result["message"] == "Path does not exist":
            os.mkdir(env_path)
        else:
            LOG.error("Kayobe environment directory %s is invalid: %s",
                      env_path, result["message"])
            sys.exit(1)

    source_config_path = parsed_args.source_config_path
    if source_config_path:
        utils.copy_dir(source_config_path, env_path, exclude=["environments"])



# Copyright (c) 2021 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ansible import errors


def get_hostvar(context, var_name, inventory_hostname=None):
    if inventory_hostname is None:
        namespace = context
    else:
        if inventory_hostname not in context['hostvars']:
            raise errors.AnsibleFilterError(
                "Inventory hostname '%s' not in hostvars" % inventory_hostname)
        namespace = context["hostvars"][inventory_hostname]
    return namespace.get(var_name)


def call_bool_filter(context, value):
    """Pass a value through the 'bool' filter.

    :param context: Jinja2 Context object.
    :param value: Value to pass through bool filter.
    :returns: A boolean.
    """
    return context.environment.call_filter("bool", value, context=context)



# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import getpass
import logging
import os
import subprocess
import sys

from kayobe import utils


LOG = logging.getLogger(__name__)

VAULT_PASSWORD_ENV = "KAYOBE_VAULT_PASSWORD"
VAULT_PASSWORD_FILE_ENV = "ANSIBLE_VAULT_PASSWORD_FILE"


def _get_vault_password_helper():
    """Return the path to the kayobe-vault-password-helper executable."""
    cmd = ["which", "kayobe-vault-password-helper"]
    try:
        # NOTE(mgoddard): universal_newlines ensures stdout is opened in text
        # mode, and we get a string rather than bytes.
        output = utils.run_command(cmd, check_output=True,
                                   universal_newlines=True)
    except subprocess.CalledProcessError:
        return None
    return output.strip()


def _get_default_vault_password_file():
    """Return the default value for the vault password file argument.

    It is possible to use an environment variable to avoid typing the vault
    password.
    """
    if (VAULT_PASSWORD_ENV not in os.environ and
            VAULT_PASSWORD_FILE_ENV not in os.environ):
        return None
    return _get_vault_password_helper()


def add_args(parser):
    """Add arguments required for running Ansible playbooks to a parser."""
    default_vault_password_file = _get_default_vault_password_file()
    vault = parser.add_mutually_exclusive_group()
    vault.add_argument("--ask-vault-pass", action="store_true",
                       help="ask for vault password")
    vault.add_argument("--vault-password-file", metavar="VAULT_PASSWORD_FILE",
                       default=default_vault_password_file,
                       help="vault password file")


def build_args(parsed_args, password_file_arg_name):
    """Build a list of command line arguments for use with ansible-playbook."""
    vault_password_file = None
    if parsed_args.ask_vault_pass:
        vault_password_file = _get_vault_password_helper()
    elif parsed_args.vault_password_file:
        vault_password_file = parsed_args.vault_password_file

    cmd = []
    if vault_password_file:
        cmd += [password_file_arg_name, vault_password_file]
    return cmd


def _validate_environment_variables():
    """Verify that only one password environment variable is set"""

    invalid_source = None
    password_env_var = None
    if VAULT_PASSWORD_ENV in os.environ:
        password_env_var = VAULT_PASSWORD_ENV
        if VAULT_PASSWORD_FILE_ENV in os.environ:
            invalid_source = "$" + VAULT_PASSWORD_FILE_ENV
    elif (VAULT_PASSWORD_FILE_ENV in os.environ):
        password_env_var = VAULT_PASSWORD_FILE_ENV
    return invalid_source, password_env_var


def _validate_args(parsed_args):
    """Verify that no conflicting arguments are being used"""

    helper = _get_vault_password_helper()
    invalid_source = None
    if parsed_args.ask_vault_pass:
        invalid_source = "--ask-vault-pass"
    elif parsed_args.vault_password_file != helper:
        invalid_source = "--vault-password-file"
    return invalid_source


def enforce_single_password_source(parsed_args):
    """Verify that a password is only being received from a single source"""
    # Ensure that a password prompt or file has not been requested if a
    # password environment variable is set, and that only one password
    # environment variable is set

    invalid_source, password_env_var = _validate_environment_variables()
    if not password_env_var:
        return

    if not invalid_source and password_env_var:
        invalid_source = _validate_args(parsed_args)

    if invalid_source:
        LOG.error("Cannot specify %s when $%s is specified" %
                  (invalid_source, password_env_var))
        sys.exit(1)


def _ask_vault_pass():
    """Prompt the user for a Vault password.

    The first time this function is called, the user is prompted for a
    password. To avoid prompting the user multiple times per invocation of
    kayobe, we cache the password and return it without prompting on subsequent
    calls.

    :return: The password entered by the user.
    """
    if not hasattr(_ask_vault_pass, "password"):
        password = getpass.getpass("Vault password: ")
        setattr(_ask_vault_pass, "password", password)
    return getattr(_ask_vault_pass, "password")


def _read_vault_password_file(vault_password_file):
    """Return the password from a vault password file."""
    vault_password = utils.read_file(vault_password_file)
    vault_password = vault_password.strip()
    return vault_password


def update_environment(parsed_args, env):
    """Update environment variables with the vault password if necessary.

    :param parsed_args: Parsed command line arguments.
    :params env: Dict of environment variables to update.
    """
    # If the Vault password has been specified via --vault-password-file, a
    # prompt has been requested via --ask-vault-pass, or the
    # $ANSIBLE_VAULT_PASSWORD_FILE environment variable is set, ensure the
    # $KAYOBE_PASSWORD_ENV environment variable is set, so that it can be
    # referenced by playbooks to generate the kolla-ansible passwords.yml
    # file.
    if VAULT_PASSWORD_ENV in env:
        return

    vault_password = None
    if VAULT_PASSWORD_FILE_ENV in os.environ:
        vault_password = _read_vault_password_file(
            os.environ[VAULT_PASSWORD_FILE_ENV])
    elif parsed_args.ask_vault_pass:
        vault_password = _ask_vault_pass()
    elif parsed_args.vault_password_file:
        vault_password = _read_vault_password_file(
            parsed_args.vault_password_file)

    if vault_password is not None:
        env[VAULT_PASSWORD_ENV] = vault_password



#!/usr/bin/env python3

# Kayobe overcloud host configure tests.
# Uses py.test and TestInfra.

import ipaddress
import os
import time

import distro
import pytest


def _is_apt():
    info = distro.linux_distribution()
    return info[0].startswith('Ubuntu')


def _is_dnf():
    info = distro.id()
    return info in ['centos', 'rocky']


def _is_dnf_mirror():
    info = distro.id()
    return info == 'centos'


def test_network_ethernet(host):
    interface = host.interface('dummy2')
    assert interface.exists
    assert '192.168.34.1' in interface.addresses
    routes = host.check_output('/sbin/ip route show dev dummy2')
    assert '192.168.40.0/24 via 192.168.34.254' in routes


def test_network_ethernet_vlan(host):
    interface = host.interface('dummy2.42')
    assert interface.exists
    assert '192.168.35.1' in interface.addresses
    assert host.file('/sys/class/net/dummy2.42/lower_dummy2').exists
    # FIXME(bbezak): remove following IF after ansible-role-interfaces
    # receive support for custom routes in NetworkManager
    if not ('centos' in host.system_info.distribution.lower() or
            'rocky' in host.system_info.distribution.lower()):
       routes = host.check_output(
           '/sbin/ip route show dev dummy2.42 table kayobe-test-route-table')
       assert '192.168.40.0/24 via 192.168.35.254' in routes
       rules = host.check_output(
           '/sbin/ip rule show table kayobe-test-route-table')
       expected = 'from 192.168.35.0/24 lookup kayobe-test-route-table'
       assert expected in rules


def test_network_bridge(host):
    interface = host.interface('br0')
    assert interface.exists
    assert '192.168.36.1' in interface.addresses
    ports = ['dummy3', 'dummy4']
    sys_ports = host.check_output('ls -1 /sys/class/net/br0/brif')
    assert sys_ports == "\n".join(ports)
    for port in ports:
       interface = host.interface(port)
       assert interface.exists
       v4_addresses = [a for a in interface.addresses
                       if ipaddress.ip_address(a).version == '4']
       assert not v4_addresses


def test_network_bridge_vlan(host):
    interface = host.interface('br0.43')
    assert interface.exists
    assert '192.168.37.1' in interface.addresses
    assert host.file('/sys/class/net/br0.43/lower_br0').exists


def test_network_bond(host):
    interface = host.interface('bond0')
    assert interface.exists
    assert '192.168.38.1' in interface.addresses
    sys_slaves = host.check_output('cat /sys/class/net/bond0/bonding/slaves')
    # Ordering is not guaranteed, so compare sets.
    sys_slaves = set(sys_slaves.split())
    slaves = set(['dummy5', 'dummy6'])
    assert sys_slaves == slaves
    for slave in slaves:
       interface = host.interface(slave)
       assert interface.exists
       assert not interface.addresses


def test_network_bond_vlan(host):
    interface = host.interface('bond0.44')
    assert interface.exists
    assert '192.168.39.1' in interface.addresses
    assert host.file('/sys/class/net/bond0.44/lower_bond0').exists


def test_network_bridge_no_ip(host):
    interface = host.interface('br1')
    assert interface.exists
    assert not '192.168.40.1' in interface.addresses


def test_additional_user_account(host):
      user = host.user("kayobe-test-user")
      assert user.name == "kayobe-test-user"
      assert user.group == "kayobe-test-user"
      assert set(user.groups) == {"kayobe-test-user", "stack"}
      assert user.gecos == "Kayobe test user"
      with host.sudo():
          assert user.password == 'kayobe-test-user-password'


def test_software_RAID(host):
    slaves = host.check_output("ls -1 /sys/class/block/md0/slaves/")
    assert slaves == "loop0\nloop1"


def test_luks(host):
    # blkid returns an emptry string without root permissions
    with host.sudo():
        blkid = host.check_output('blkid /dev/md0')
    assert 'TYPE="crypto_LUKS"' in blkid


def test_sysctls(host):
    assert host.sysctl("fs.mount-max") == 99999


def test_cloud_init_is_disabled(host):
    assert host.file("/etc/cloud/cloud-init.disabled").exists


def test_docker_storage_driver_is_devicemapper(host):
    with host.sudo("stack"):
        info = host.check_output("docker info")
    assert "devicemapper" in info


@pytest.mark.parametrize('user', ['kolla', 'stack'])
def test_docker_image_download(host, user):
    with host.sudo(user):
        host.check_output("docker pull alpine")


@pytest.mark.parametrize('user', ['kolla', 'stack'])
def test_docker_container_run(host, user):
    with host.sudo(user):
        host.check_output("docker run --rm alpine /bin/true")


def test_timezone(host):
    status = host.check_output("timedatectl status")
    assert "Pacific/Honolulu" in status


def test_ntp_alternative_services_disabled(host):
    # Tests that we don't have any conflicting NTP servers running
    # NOTE(wszumski): We always mask services even if they don't exist
    ntpd_service = host.service("ntp")
    assert ntpd_service.is_masked
    assert not ntpd_service.is_running

    timesyncd_service = host.service("systemd-timesyncd")
    assert timesyncd_service.is_masked
    assert not timesyncd_service.is_running


def test_ntp_running(host):
    # Tests that NTP services are enabled and running
    assert host.package("chrony").is_installed
    assert host.service("chronyd").is_enabled
    assert host.service("chronyd").is_running


def test_ntp_non_default_time_server(host):
    # Tests that the NTP pool has been changed from pool.ntp.org to
    # time.cloudflare.com
    if ('centos' in host.system_info.distribution.lower() or
       'rocky' in host.system_info.distribution.lower()):
        chrony_config = host.file("/etc/chrony.conf")
    else:
        # Debian based distributions use the following path
        chrony_config = host.file("/etc/chrony/chrony.conf")
    assert chrony_config.exists
    assert "time.cloudflare.com" in chrony_config.content_string


def test_ntp_clock_synchronized(host):
    # Tests that the clock is synchronized
    status_output = host.check_output("timedatectl status")
    assert "synchronized: yes" in status_output


@pytest.mark.skipif(not _is_apt(), reason="Apt only supported on Ubuntu")
def test_apt_config(host):
    apt_config = host.file("/etc/apt/apt.conf.d/99retries")
    assert apt_config.exists
    assert apt_config.content_string == "Acquire::Retries 1;\n"


@pytest.mark.skipif(not _is_apt(), reason="Apt only supported on Ubuntu")
def test_apt_custom_package_repository_is_available(host):
    with host.sudo():
        host.check_output("apt -y install td-agent")
    assert host.package("td-agent").is_installed


@pytest.mark.parametrize('repo', ["appstream", "baseos", "extras", "epel"])
@pytest.mark.skipif(not _is_dnf_mirror(), reason="DNF OpenDev mirror only for CentOS 8")
def test_dnf_local_package_mirrors(host, repo):
    # Depends on SITE_MIRROR_FQDN environment variable.
    assert os.getenv('SITE_MIRROR_FQDN')
    # NOTE(mgoddard): Should not require sudo but some files
    # (/var/cache/dnf/expired_repos.json) can have incorrect permissions.
    # https://bugzilla.redhat.com/show_bug.cgi?id=1636909
    with host.sudo():
        info = host.check_output("dnf repoinfo %s", repo)
    assert os.getenv('SITE_MIRROR_FQDN') in info


@pytest.mark.skipif(not _is_dnf(), reason="DNF only supported on CentOS/Rocky")
def test_dnf_custom_package_repository_is_available(host):
    with host.sudo():
        host.check_output("dnf -y install td-agent")
    assert host.package("td-agent").is_installed


@pytest.mark.skipif(not _is_dnf(), reason="DNF only supported on CentOS/Rocky")
def test_dnf_automatic(host):
    assert host.package("dnf-automatic").is_installed
    assert host.service("dnf-automatic.timer").is_enabled
    assert host.service("dnf-automatic.timer").is_running


@pytest.mark.skipif(not _is_dnf(),
                    reason="tuned profile setting only supported on CentOS/Rocky")
def test_tuned_profile_is_active(host):
    tuned_output = host.check_output("tuned-adm active")
    assert "throughput-performance" in tuned_output


def test_firewalld_running(host):
    assert host.package("firewalld").is_installed
    assert host.service("firewalld.service").is_enabled
    assert host.service("firewalld.service").is_running


def test_firewalld_zones(host):
    # Verify that interfaces are on correct zones.
    expected_zones = {
        'dummy2.42': 'test-zone1',
        'br0': 'test-zone2',
        'br0.43': 'test-zone3',
        'bond0': 'test-zone3',
        'bond0.44': 'public'
    }
    for interface, expected_zone in expected_zones.items():
        with host.sudo():
            zone = host.check_output(
                "firewall-cmd --get-zone-of-interface %s", interface)
            assert zone == expected_zone

            zone = host.check_output(
                "firewall-cmd --permanent --get-zone-of-interface %s",
                interface)
            assert zone == expected_zone


def test_firewalld_rules(host):
    # Verify that expected rules are present.
    expected_info = {
        'test-zone1': [
            '  services: ',
            '  ports: 8080/tcp',
            '  icmp-blocks: ',
        ],
        'test-zone2': [
            '  services: http',
            '  ports: ',
            '  icmp-blocks: ',
        ],
        'test-zone3': [
            '  services: ',
            '  ports: ',
            '  icmp-blocks: echo-request',
        ],
        'public': [
            '  services: dhcpv6-client ssh',
            '  ports: ',
            '  icmp-blocks: ',
        ],
    }

    for zone, expected_lines in expected_info.items():
        with host.sudo():
            info = host.check_output(
                "firewall-cmd --info-zone %s", zone)
            info = info.splitlines()
            perm_info = host.check_output(
                "firewall-cmd --permanent --info-zone %s", zone)
            perm_info = perm_info.splitlines()

        for expected_line in expected_lines:
            assert expected_line in info
            assert expected_line in perm_info


def test_swap(host):
    swapon = host.check_output("swapon -s")
    swapon = swapon.splitlines()
    assert len(swapon) > 1
    swap_devs = [swap.split()[0] for swap in swapon[1:]]
    assert "/swapfile" in swap_devs



# Copyright (c) 2021 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import unittest

from ansible import errors
from ansible.plugins.filter.core import to_bool
import jinja2

from kayobe.plugins.filter import networkd


class BaseNetworkdTest(unittest.TestCase):

    maxDiff = 2000

    variables = {
        # Inventory hostname, used to index IP list.
        "inventory_hostname": "test-host",
        # net1: Ethernet on eth0 with IP 1.2.3.4/24.
        "net1_interface": "eth0",
        "net1_cidr": "1.2.3.0/24",
        "net1_ips": {"test-host": "1.2.3.4"},
        # net2: VLAN on eth0.2 with VLAN 2 on interface eth0.
        "net2_interface": "eth0.2",
        "net2_vlan": 2,
        # net3: bridge on br0 with ports eth0 and eth1.
        "net3_interface": "br0",
        "net3_bridge_ports": ["eth0", "eth1"],
        # net4: bond on bond0 with members eth0 and eth1.
        "net4_interface": "bond0",
        "net4_bond_slaves": ["eth0", "eth1"],
        # Prefix for networkd config file names.
        "networkd_prefix": "50-kayobe-",
        # Veth pair patch link prefix and suffix.
        "network_patch_prefix": "p-",
        "network_patch_suffix_ovs": "-ovs",
        "network_patch_suffix_phy": "-phy",
        # List of route tables.
        "network_route_tables": [],
    }

    def setUp(self):
        # Bandit complains about Jinja2 autoescaping without nosec.
        self.env = jinja2.Environment()  # nosec
        self.env.filters['bool'] = to_bool
        self.context = self._make_context(self.variables)

    def _make_context(self, parent):
        return self.env.context_class(
            self.env, parent=parent, name='dummy', blocks={})

    def _update_context(self, variables):
        updated_vars = copy.deepcopy(self.variables)
        updated_vars.update(variables)
        self.context = self._make_context(updated_vars)


class TestNetworkdNetDevs(BaseNetworkdTest):

    def test_empty(self):
        devs = networkd.networkd_netdevs(self.context, [])
        self.assertEqual({}, devs)

    def test_eth(self):
        devs = networkd.networkd_netdevs(self.context, ["net1"])
        expected = {}
        self.assertEqual(expected, devs)

    def test_eth_untagged_vlan(self):
        # An untagged interface on a network with a VLAN defined should not
        # create a VLAN subinterface.
        self._update_context({"net1_vlan": 42})
        devs = networkd.networkd_netdevs(self.context, ["net1"])
        expected = {}
        self.assertEqual(expected, devs)

    def test_vlan(self):
        devs = networkd.networkd_netdevs(self.context, ["net2"])
        expected = {
            "50-kayobe-eth0.2": [
                {
                    "NetDev": [
                        {"Name": "eth0.2"},
                        {"Kind": "vlan"},
                    ]
                },
                {
                    "VLAN": [
                        {"Id": 2},
                    ]
                },
            ]
        }
        self.assertEqual(expected, devs)

    def test_vlan_all_options(self):
        self._update_context({"net2_mtu": 1400})
        devs = networkd.networkd_netdevs(self.context, ["net2"])
        expected = {
            "50-kayobe-eth0.2": [
                {
                    "NetDev": [
                        {"Name": "eth0.2"},
                        {"Kind": "vlan"},
                        {"MTUBytes": 1400},
                    ]
                },
                {
                    "VLAN": [
                        {"Id": 2},
                    ]
                },
            ]
        }
        self.assertEqual(expected, devs)

    def test_vlan_no_interface(self):
        self._update_context({"net2_interface": None})
        self.assertRaises(errors.AnsibleFilterError,
                          networkd.networkd_netdevs, self.context, ["net2"])

    def test_bridge(self):
        devs = networkd.networkd_netdevs(self.context, ["net3"])
        expected = {
            "50-kayobe-br0": [
                {
                    "NetDev": [
                        {"Name": "br0"},
                        {"Kind": "bridge"},
                    ]
                },
            ]
        }
        self.assertEqual(expected, devs)

    def test_bridge_all_options(self):
        self._update_context({"net3_mtu": 1400})
        devs = networkd.networkd_netdevs(self.context, ["net3"])
        expected = {
            "50-kayobe-br0": [
                {
                    "NetDev": [
                        {"Name": "br0"},
                        {"Kind": "bridge"},
                        {"MTUBytes": 1400},
                    ]
                },
            ]
        }
        self.assertEqual(expected, devs)

    def test_bridge_no_interface(self):
        self._update_context({"net3_interface": None})
        self.assertRaises(errors.AnsibleFilterError,
                          networkd.networkd_netdevs, self.context, ["net3"])

    def test_bridge_untagged_vlan(self):
        # A bridge on a network with a VLAN defined should not create a VLAN
        # subinterface.
        self._update_context({"net3_vlan": 42})
        devs = networkd.networkd_netdevs(self.context, ["net3"])
        expected = {
            "50-kayobe-br0": [
                {
                    "NetDev": [
                        {"Name": "br0"},
                        {"Kind": "bridge"},
                    ]
                },
            ]
        }
        self.assertEqual(expected, devs)

    def test_bond(self):
        devs = networkd.networkd_netdevs(self.context, ["net4"])
        expected = {
            "50-kayobe-bond0": [
                {
                    "NetDev": [
                        {"Name": "bond0"},
                        {"Kind": "bond"},
                    ]
                },
            ]
        }
        self.assertEqual(expected, devs)

    def test_bond_all_options(self):
        self._update_context({
            "net4_mtu": 1400,
            "net4_bond_mode": "802.3ad",
            "net4_bond_miimon": 100,
            "net4_bond_updelay": 200,
            "net4_bond_downdelay": 300,
            "net4_bond_xmit_hash_policy": "layer3+4",
            "net4_bond_lacp_rate": 60,
        })
        devs = networkd.networkd_netdevs(self.context, ["net4"])
        expected = {
            "50-kayobe-bond0": [
                {
                    "NetDev": [
                        {"Name": "bond0"},
                        {"Kind": "bond"},
                        {"MTUBytes": 1400},
                    ]
                },
                {
                    "Bond": [
                        {"Mode": "802.3ad"},
                        {"TransmitHashPolicy": "layer3+4"},
                        {"LACPTransmitRate": 60},
                        {"MIIMonitorSec": 0.1},
                        {"UpDelaySec": 0.2},
                        {"DownDelaySec": 0.3},
                    ]
                },
            ]
        }
        self.assertEqual(expected, devs)

    def test_bond_no_interface(self):
        self._update_context({"net4_interface": None})
        self.assertRaises(errors.AnsibleFilterError,
                          networkd.networkd_netdevs, self.context, ["net4"])

    def test_bond_untagged_vlan(self):
        # A bridge on a network with a VLAN defined should not create a VLAN
        # subinterface.
        self._update_context({"net4_vlan": 42})
        devs = networkd.networkd_netdevs(self.context, ["net4"])
        expected = {
            "50-kayobe-bond0": [
                {
                    "NetDev": [
                        {"Name": "bond0"},
                        {"Kind": "bond"},
                    ]
                },
            ]
        }
        self.assertEqual(expected, devs)

    def test_veth(self):
        self._update_context({"external_net_names": ["net3"]})
        devs = networkd.networkd_netdevs(self.context, ["net3"])
        expected = {
            "50-kayobe-br0": [
                {
                    "NetDev": [
                        {"Name": "br0"},
                        {"Kind": "bridge"},
                    ]
                },
            ],
            "50-kayobe-p-br0-phy": [
                {
                    "NetDev": [
                        {"Name": "p-br0-phy"},
                        {"Kind": "veth"},
                    ]
                },
                {
                    "Peer": [
                        {"Name": "p-br0-ovs"},
                    ]
                },
            ]
        }
        self.assertEqual(expected, devs)

    def test_veth_no_interface(self):
        self._update_context({"external_net_names": ["net3"],
                              "net3_interface": None})
        self.assertRaises(errors.AnsibleFilterError,
                          networkd.networkd_netdevs, self.context, ["net3"])


class TestNetworkdLinks(BaseNetworkdTest):

    def test_empty(self):
        links = networkd.networkd_links(self.context, ['net1'])
        self.assertEqual({}, links)


class TestNetworkdNetworks(BaseNetworkdTest):

    def test_empty(self):
        nets = networkd.networkd_networks(self.context, [])
        self.assertEqual({}, nets)

    def test_eth(self):
        nets = networkd.networkd_networks(self.context, ["net1"])
        expected = {
            "50-kayobe-eth0": [
                {
                    "Match": [
                        {"Name": "eth0"}
                    ]
                },
                {
                    "Network": [
                        {"Address": "1.2.3.4/24"},
                    ]
                },
            ]
        }
        self.assertEqual(expected, nets)

    def test_eth_all_options(self):
        self._update_context({
            "net1_gateway": "1.2.3.1",
            "net1_mtu": 1400,
            "net1_routes": [
                {
                    "cidr": "1.2.4.0/24",
                },
                {
                    "cidr": "1.2.5.0/24",
                    "gateway": "1.2.5.1",
                },
                {
                    "cidr": "1.2.6.0/24",
                    "table": 42,
                },
            ],
            "net1_rules": [
                {
                    "from": "1.2.7.0/24",
                    "table": 43,
                },
                {
                    "to": "1.2.8.0/24",
                    "table": 44,
                    "priority": 1,
                },
            ],
            "net1_bootproto": "dhcp",
            "net1_defroute": 'no',
        })
        nets = networkd.networkd_networks(self.context, ["net1"])
        expected = {
            "50-kayobe-eth0": [
                {
                    "Match": [
                        {"Name": "eth0"}
                    ]
                },
                {
                    "Network": [
                        {"Address": "1.2.3.4/24"},
                        {"Gateway": "1.2.3.1"},
                        {"DHCP": "yes"},
                        {'UseGateway': "false"},
                    ]
                },
                {
                    "Link": [
                        {"MTUBytes": 1400},
                    ]
                },
                {
                    "Route": [
                        {"Destination": "1.2.4.0/24"},
                    ]
                },
                {
                    "Route": [
                        {"Destination": "1.2.5.0/24"},
                        {"Gateway": "1.2.5.1"},
                    ]
                },
                {
                    "Route": [
                        {"Destination": "1.2.6.0/24"},
                        {"Table": 42},
                    ]
                },
                {
                    "RoutingPolicyRule": [
                        {"From": "1.2.7.0/24"},
                        {"Table": 43},
                    ]
                },
                {
                    "RoutingPolicyRule": [
                        {"To": "1.2.8.0/24"},
                        {"Priority": 1},
                        {"Table": 44},
                    ]
                },
            ]
        }
        self.assertEqual(expected, nets)

    def test_eth_no_interface(self):
        self._update_context({"net1_interface": None})
        self.assertRaises(errors.AnsibleFilterError,
                          networkd.networkd_networks, self.context, ["net1"])

    def test_vlan(self):
        nets = networkd.networkd_networks(self.context, ["net2"])
        expected = {
            "50-kayobe-eth0": [
                {
                    "Match": [
                        {"Name": "eth0"}
                    ],
                },
                {
                    "Network": [
                        {"VLAN": "eth0.2"},
                    ]
                },
            ],
            "50-kayobe-eth0.2": [
                {
                    "Match": [
                        {"Name": "eth0.2"}
                    ]
                },
            ]
        }
        self.assertEqual(expected, nets)

    def test_vlan_with_parent(self):
        nets = networkd.networkd_networks(self.context, ["net1", "net2"])
        expected = {
            "50-kayobe-eth0": [
                {
                    "Match": [
                        {"Name": "eth0"}
                    ]
                },
                {
                    "Network": [
                        {"Address": "1.2.3.4/24"},
                        {"VLAN": "eth0.2"},
                    ]
                },
            ],
            "50-kayobe-eth0.2": [
                {
                    "Match": [
                        {"Name": "eth0.2"}
                    ]
                },
            ]
        }
        self.assertEqual(expected, nets)

    def test_vlan_no_interface(self):
        self._update_context({"net2_interface": None})
        self.assertRaises(errors.AnsibleFilterError,
                          networkd.networkd_networks, self.context, ["net2"])

    def test_bridge(self):
        nets = networkd.networkd_networks(self.context, ["net3"])
        expected = {
            "50-kayobe-br0": [
                {
                    "Match": [
                        {"Name": "br0"}
                    ]
                },
            ],
            "50-kayobe-eth0": [
                {
                    "Match": [
                        {"Name": "eth0"}
                    ]
                },
                {
                    "Network": [
                        {"Bridge": "br0"},
                    ]
                },
            ],
            "50-kayobe-eth1": [
                {
                    "Match": [
                        {"Name": "eth1"}
                    ]
                },
                {
                    "Network": [
                        {"Bridge": "br0"},
                    ]
                },
            ]
        }
        self.assertEqual(expected, nets)

    def test_bridge_with_bridge_port_net(self):
        # Test the case where a bridge port interface is a Kayobe network
        # (here, eth0 is net1).
        self._update_context({
            "net1_mtu": 1400,
            "net1_ips": None,
        })
        nets = networkd.networkd_networks(self.context, ["net1", "net3"])
        expected = {
            "50-kayobe-br0": [
                {
                    "Match": [
                        {"Name": "br0"}
                    ]
                },
            ],
            "50-kayobe-eth0": [
                {
                    "Match": [
                        {"Name": "eth0"}
                    ]
                },
                {
                    "Network": [
                        {"Bridge": "br0"},
                    ]
                },
                {
                    "Link": [
                        {"MTUBytes": 1400},
                    ]
                },
            ],
            "50-kayobe-eth1": [
                {
                    "Match": [
                        {"Name": "eth1"}
                    ]
                },
                {
                    "Network": [
                        {"Bridge": "br0"},
                    ]
                },
            ]
        }
        self.assertEqual(expected, nets)

    def test_bridge_with_bridge_port_vlan(self):
        # Test the case where one of the bridge ports has a VLAN subinterface.
        self._update_context({
            "net2_interface": "eth1.2",
        })
        nets = networkd.networkd_networks(self.context, ["net2", "net3"])
        expected = {
            "50-kayobe-br0": [
                {
                    "Match": [
                        {"Name": "br0"}
                    ]
                },
            ],
            "50-kayobe-eth0": [
                {
                    "Match": [
                        {"Name": "eth0"}
                    ]
                },
                {
                    "Network": [
                        {"Bridge": "br0"},
                    ]
                },
            ],
            "50-kayobe-eth1": [
                {
                    "Match": [
                        {"Name": "eth1"}
                    ]
                },
                {
                    "Network": [
                        {"Bridge": "br0"},
                        {"VLAN": "eth1.2"}
                    ]
                },
            ],
            "50-kayobe-eth1.2": [
                {
                    "Match": [
                        {"Name": "eth1.2"}
                    ]
                },
            ],
        }
        self.assertEqual(expected, nets)

    def test_bridge_with_bridge_port_vlan_net(self):
        # Test the case where one of the bridge ports has a VLAN subinterface,
        # and is also a Kayobe network (here eth0 is net1).
        self._update_context({
            "net1_ips": None,
        })
        nets = networkd.networkd_networks(self.context,
                                          ["net1", "net2", "net3"])
        expected = {
            "50-kayobe-br0": [
                {
                    "Match": [
                        {"Name": "br0"}
                    ]
                },
            ],
            "50-kayobe-eth0": [
                {
                    "Match": [
                        {"Name": "eth0"}
                    ]
                },
                {
                    "Network": [
                        {"Bridge": "br0"},
                        {"VLAN": "eth0.2"}
                    ]
                },
            ],
            "50-kayobe-eth0.2": [
                {
                    "Match": [
                        {"Name": "eth0.2"}
                    ]
                },
            ],
            "50-kayobe-eth1": [
                {
                    "Match": [
                        {"Name": "eth1"}
                    ]
                },
                {
                    "Network": [
                        {"Bridge": "br0"},
                    ]
                },
            ]
        }
        self.assertEqual(expected, nets)

    def test_bridge_no_interface(self):
        self._update_context({"net3_interface": None})
        self.assertRaises(errors.AnsibleFilterError,
                          networkd.networkd_networks, self.context, ["net3"])

    def test_bond(self):
        nets = networkd.networkd_networks(self.context, ["net4"])
        expected = {
            "50-kayobe-bond0": [
                {
                    "Match": [
                        {"Name": "bond0"}
                    ]
                },
            ],
            "50-kayobe-eth0": [
                {
                    "Match": [
                        {"Name": "eth0"}
                    ]
                },
                {
                    "Network": [
                        {"Bond": "bond0"},
                    ]
                },
            ],
            "50-kayobe-eth1": [
                {
                    "Match": [
                        {"Name": "eth1"}
                    ]
                },
                {
                    "Network": [
                        {"Bond": "bond0"},
                    ]
                },
            ]
        }
        self.assertEqual(expected, nets)

    def test_bond_with_bond_member_net(self):
        # Test the case where a bond member interface is a Kayobe network
        # (here, eth0 is net1).
        self._update_context({
            "net1_mtu": 1400,
            "net1_ips": None,
        })
        nets = networkd.networkd_networks(self.context, ["net1", "net4"])
        expected = {
            "50-kayobe-bond0": [
                {
                    "Match": [
                        {"Name": "bond0"}
                    ]
                },
            ],
            "50-kayobe-eth0": [
                {
                    "Match": [
                        {"Name": "eth0"}
                    ]
                },
                {
                    "Network": [
                        {"Bond": "bond0"},
                    ]
                },
                {
                    "Link": [
                        {"MTUBytes": 1400},
                    ]
                },
            ],
            "50-kayobe-eth1": [
                {
                    "Match": [
                        {"Name": "eth1"}
                    ]
                },
                {
                    "Network": [
                        {"Bond": "bond0"},
                    ]
                },
            ]
        }
        self.assertEqual(expected, nets)

    def test_bond_with_bond_member_vlan(self):
        # Test the case where one of the bond members has a VLAN subinterface.
        self._update_context({
            "net2_interface": "eth1.2",
        })
        nets = networkd.networkd_networks(self.context, ["net2", "net4"])
        expected = {
            "50-kayobe-bond0": [
                {
                    "Match": [
                        {"Name": "bond0"}
                    ]
                },
            ],
            "50-kayobe-eth0": [
                {
                    "Match": [
                        {"Name": "eth0"}
                    ]
                },
                {
                    "Network": [
                        {"Bond": "bond0"},
                    ]
                },
            ],
            "50-kayobe-eth1": [
                {
                    "Match": [
                        {"Name": "eth1"}
                    ]
                },
                {
                    "Network": [
                        {"Bond": "bond0"},
                        {"VLAN": "eth1.2"},
                    ]
                },
            ],
            "50-kayobe-eth1.2": [
                {
                    "Match": [
                        {"Name": "eth1.2"}
                    ]
                },
            ],
        }
        self.assertEqual(expected, nets)

    def test_bond_with_bond_member_vlan_net(self):
        # Test the case where one of the bond members has a VLAN subinterface,
        # and is also a Kayobe network (here eth0 is net1).
        self._update_context({
            "net1_ips": None,
        })
        nets = networkd.networkd_networks(self.context,
                                          ["net1", "net2", "net4"])
        expected = {
            "50-kayobe-bond0": [
                {
                    "Match": [
                        {"Name": "bond0"}
                    ]
                },
            ],
            "50-kayobe-eth0": [
                {
                    "Match": [
                        {"Name": "eth0"}
                    ]
                },
                {
                    "Network": [
                        {"Bond": "bond0"},
                        {"VLAN": "eth0.2"},
                    ]
                },
            ],
            "50-kayobe-eth0.2": [
                {
                    "Match": [
                        {"Name": "eth0.2"}
                    ]
                },
            ],
            "50-kayobe-eth1": [
                {
                    "Match": [
                        {"Name": "eth1"}
                    ]
                },
                {
                    "Network": [
                        {"Bond": "bond0"},
                    ]
                },
            ]
        }
        self.assertEqual(expected, nets)

    def test_bond_no_interface(self):
        self._update_context({"net4_interface": None})
        self.assertRaises(errors.AnsibleFilterError,
                          networkd.networkd_networks, self.context, ["net4"])

    def test_veth(self):
        self._update_context({"external_net_names": ["net3"],
                              "net3_bridge_ports": []})
        nets = networkd.networkd_networks(self.context, ["net3"])
        expected = {
            "50-kayobe-br0": [
                {
                    "Match": [
                        {"Name": "br0"}
                    ]
                },
            ],
            "50-kayobe-p-br0-phy": [
                {
                    "Match": [
                        {"Name": "p-br0-phy"}
                    ]
                },
                {
                    "Network": [
                        {"Bridge": "br0"},
                    ]
                },
            ],
            "50-kayobe-p-br0-ovs": [
                {
                    "Match": [
                        {"Name": "p-br0-ovs"}
                    ]
                },
                {
                    "Network": [
                        {"ConfigureWithoutCarrier": "true"},
                    ]
                },
            ],
        }
        self.assertEqual(expected, nets)

    def test_veth_with_mtu(self):
        self._update_context({"external_net_names": ["net3"],
                              "net3_bridge_ports": [],
                              "net3_mtu": 1400})
        nets = networkd.networkd_networks(self.context, ["net3"])
        expected = {
            "50-kayobe-br0": [
                {
                    "Match": [
                        {"Name": "br0"}
                    ]
                },
                {
                    "Link": [
                        {"MTUBytes": 1400},
                    ]
                },
            ],
            "50-kayobe-p-br0-phy": [
                {
                    "Match": [
                        {"Name": "p-br0-phy"}
                    ]
                },
                {
                    "Network": [
                        {"Bridge": "br0"},
                    ]
                },
                {
                    "Link": [
                        {"MTUBytes": 1400},
                    ]
                },
            ],
            "50-kayobe-p-br0-ovs": [
                {
                    "Match": [
                        {"Name": "p-br0-ovs"}
                    ]
                },
                {
                    "Network": [
                        {"ConfigureWithoutCarrier": "true"},
                    ]
                },
                {
                    "Link": [
                        {"MTUBytes": 1400},
                    ]
                },
            ],
        }
        self.assertEqual(expected, nets)

    def test_veth_on_vlan(self):
        # Test the case where a VLAN interface is one of the networks that
        # needs patching to OVS. The parent interface is a bridge, and the veth
        # pair should be plugged into it.
        self._update_context({
            "provision_wl_net_name": "net5",
            "net3_bridge_ports": [],
            "net5_interface": "br0.42",
            "net5_vlan": 42})
        nets = networkd.networkd_networks(self.context, ["net3", "net5"])
        expected = {
            "50-kayobe-br0": [
                {
                    "Match": [
                        {"Name": "br0"}
                    ]
                },
                {
                    "Network": [
                        {"VLAN": "br0.42"}
                    ]
                }
            ],
            "50-kayobe-br0.42": [
                {
                    "Match": [
                        {"Name": "br0.42"}
                    ]
                },
            ],
            "50-kayobe-p-br0-phy": [
                {
                    "Match": [
                        {"Name": "p-br0-phy"}
                    ]
                },
                {
                    "Network": [
                        {"Bridge": "br0"},
                    ]
                },
            ],
            "50-kayobe-p-br0-ovs": [
                {
                    "Match": [
                        {"Name": "p-br0-ovs"}
                    ]
                },
                {
                    "Network": [
                        {"ConfigureWithoutCarrier": "true"},
                    ]
                },
            ],
        }
        self.assertEqual(expected, nets)

    def test_veth_no_interface(self):
        self._update_context({"external_net_names": ["net3"],
                              "net3_interface": None})
        self.assertRaises(errors.AnsibleFilterError,
                          networkd.networkd_networks, self.context, ["net3"])

    def test_no_veth_without_bridge(self):
        self._update_context({"external_net_names": ["net1"]})
        nets = networkd.networkd_networks(self.context, ["net1"])
        expected = {
            "50-kayobe-eth0": [
                {
                    "Match": [
                        {"Name": "eth0"}
                    ]
                },
                {
                    "Network": [
                        {"Address": "1.2.3.4/24"},
                    ]
                },
            ]
        }
        self.assertEqual(expected, nets)

    def test_no_veth_on_vlan_without_bridge(self):
        # Test the case where a VLAN interface is one of the networks that
        # needs patching to OVS. The parent interface is a bridge, and the veth
        # pair should be plugged into it.
        self._update_context({"provision_wl_net": "net2"})
        nets = networkd.networkd_networks(self.context, ["net1", "net2"])
        expected = {
            "50-kayobe-eth0": [
                {
                    "Match": [
                        {"Name": "eth0"}
                    ]
                },
                {
                    "Network": [
                        {"Address": "1.2.3.4/24"},
                        {"VLAN": "eth0.2"},
                    ]
                },
            ],
            "50-kayobe-eth0.2": [
                {
                    "Match": [
                        {"Name": "eth0.2"}
                    ]
                },
            ]
        }
        self.assertEqual(expected, nets)



# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import sys

from cliff.app import App
from cliff.commandmanager import CommandManager

from kayobe import version


class KayobeApp(App):

    def __init__(self):
        release_version = version.version_info.release_string()
        super(KayobeApp, self).__init__(
            description='Kayobe Command Line Interface (CLI)',
            version=release_version,
            command_manager=CommandManager('kayobe.cli'),
            deferred_help=True,
            )

    def initialize_app(self, argv):
        self.LOG.debug('initialize_app')

    def prepare_to_run_command(self, cmd):
        self.LOG.debug('prepare_to_run_command %s', cmd.__class__.__name__)

    def clean_up(self, cmd, result, err):
        self.LOG.debug('clean_up %s', cmd.__class__.__name__)
        if err:
            self.LOG.debug('got an error: %s', err)


def main(argv=sys.argv[1:]):
    myapp = KayobeApp()
    return myapp.run(argv)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



# Copyright 2015 Sam Yaple
# Copyright 2016 intel
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file has been adapted from the merge_yaml action plugin in Kolla
# Ansible.
# https://opendev.org/openstack/kolla-ansible/src/branch/master/ansible/action_plugins/merge_yaml.py

import os
import shutil
import tempfile

import yaml

from ansible import constants
from ansible import errors as ansible_errors
from ansible.plugins import action

DOCUMENTATION = '''
---
module: merge_yaml
short_description: Merge yaml-style configs
description:
     - PyYAML is used to merge several yaml files into one
options:
  dest:
    description:
      - The destination file name
    required: True
    type: str
  sources:
    description:
      - A list of files on the destination node to merge together
    default: None
    required: True
    type: str
  extend_lists:
    description:
      - For a given key referencing a list, this determines whether
        the list items should be combined with the items in another
        document if an equivalent key is found. An equivalent key
        has the same parents and value as the first. The default
        behaviour is to replace existing entries i.e if you have
        two yaml documents that both define a list with an equivalent
        key, the value from the document that appears later in the
        list of sources will replace the value that appeared in the
        earlier one.
    default: False
    required: False
    type: bool
author: Sean Mooney
'''

EXAMPLES = '''
Merge multiple yaml files:

- hosts: localhost
  tasks:
    - name: Merge yaml files
      merge_yaml:
        sources:
          - "/tmp/default.yml"
          - "/tmp/override.yml"
        dest:
          - "/tmp/out.yml"
'''


class ActionModule(action.ActionBase):

    TRANSFERS_FILES = True

    def read_config(self, source):
        result = None
        # Only use config if present
        if source and os.access(source, os.R_OK):
            with open(source, 'r') as f:
                template_data = f.read()

            # set search path to mimic 'template' module behavior
            searchpath = [
                self._loader._basedir,
                os.path.join(self._loader._basedir, 'templates'),
                os.path.dirname(source),
            ]
            self._templar.environment.loader.searchpath = searchpath

            template_data = self._templar.template(template_data)
            result = yaml.safe_load(template_data)
        return result or {}

    def run(self, tmp=None, task_vars=None):
        if task_vars is None:
            task_vars = dict()
        result = super(ActionModule, self).run(tmp, task_vars)
        del tmp  # not used

        # save template args.
        extra_vars = self._task.args.get('vars', list())
        old_vars = self._templar._available_variables

        temp_vars = task_vars.copy()
        temp_vars.update(extra_vars)
        self._templar.available_variables = temp_vars

        output = {}
        sources = self._task.args.get('sources', None)
        extend_lists = self._task.args.get('extend_lists', False)
        if not isinstance(sources, list):
            sources = [sources]
        for source in sources:
            Utils.update_nested_conf(
                output, self.read_config(source), extend_lists)

        # restore original vars
        self._templar.available_variables = old_vars

        local_tempdir = tempfile.mkdtemp(dir=constants.DEFAULT_LOCAL_TMP)

        try:
            result_file = os.path.join(local_tempdir, 'source')
            with open(result_file, 'w') as f:
                f.write(yaml.dump(output, default_flow_style=False))

            new_task = self._task.copy()
            new_task.args.pop('sources', None)
            new_task.args.pop('extend_lists', None)
            new_task.args.update(
                dict(
                    src=result_file
                )
            )

            copy_action = self._shared_loader_obj.action_loader.get(
                'copy',
                task=new_task,
                connection=self._connection,
                play_context=self._play_context,
                loader=self._loader,
                templar=self._templar,
                shared_loader_obj=self._shared_loader_obj)
            result.update(copy_action.run(task_vars=task_vars))
        finally:
            shutil.rmtree(local_tempdir)
        return result


class Utils(object):
    @staticmethod
    def update_nested_conf(conf, update, extend_lists=False):
        for k, v in update.items():
            if isinstance(v, dict):
                conf[k] = Utils.update_nested_conf(
                    conf.get(k, {}), v, extend_lists)
            elif k in conf and isinstance(conf[k], list) and extend_lists:
                if not isinstance(v, list):
                    errmsg = (
                        "Failure merging key `%(key)s` in dictionary "
                        "`%(dictionary)s`. Expecting a list, but received: "
                        "`%(value)s`, which is of type: `%(type)s`" % {
                            "key": k, "dictionary": conf,
                            "value": v, "type": type(v)}
                    )
                    raise ansible_errors.AnsibleModuleError(errmsg)
                conf[k].extend(v)
            else:
                conf[k] = v
        return conf



# Copyright (c) 2021 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import os
import unittest
from unittest import mock

from kayobe import ansible
from kayobe import environment
from kayobe import utils


class TestCase(unittest.TestCase):

    maxDiff = None

    @mock.patch.object(utils, "is_readable_dir")
    def test_unreadable_environments_directory(self, mock_readable_dir):
        mock_readable_dir.return_value = {
            "result": False,
            "message": "Directory is not readable"
        }
        parser = argparse.ArgumentParser()
        args = [
            "--config-path", "/path/to/config",
            "--environment", "foo",
        ]
        ansible.add_args(parser)
        environment.add_args(parser)
        parsed_args = parser.parse_args(args)
        self.assertRaises(SystemExit,
                          environment.create_kayobe_environment, parsed_args)

    @mock.patch.object(utils, "is_readable_dir")
    def test_environment_exists(self, mock_readable_dir):
        mock_readable_dir.side_effect = [{"result": True}, {"result": True}]
        parser = argparse.ArgumentParser()
        args = [
            "--config-path", "/path/to/config",
            "--environment", "foo",
        ]
        ansible.add_args(parser)
        environment.add_args(parser)
        parsed_args = parser.parse_args(args)
        self.assertRaises(SystemExit,
                          environment.create_kayobe_environment, parsed_args)

    @mock.patch.object(utils, "copy_dir")
    @mock.patch.object(os, "mkdir")
    @mock.patch.object(utils, "is_readable_dir")
    def test_create_kayobe_environment(self, mock_readable_dir, mock_mkdir,
                                       mock_copy_dir):
        mock_readable_dir.return_value = {
            "result": False,
            "message": "Path does not exist"
        }
        parser = argparse.ArgumentParser()
        args = [
            "--config-path", "/path/to/config",
            "--source-config-path", "/path/to/foo",
            "--environment", "foo",
        ]
        ansible.add_args(parser)
        environment.add_args(parser)
        parsed_args = parser.parse_args(args)
        environment.create_kayobe_environment(parsed_args)
        expected_calls = [
            mock.call("/path/to/config/environments"),
            mock.call("/path/to/config/environments/foo"),
        ]
        self.assertListEqual(expected_calls, mock_mkdir.call_args_list)
        mock_copy_dir.assert_called_once_with(
            "/path/to/foo", "/path/to/config/environments/foo",
            exclude=["environments"])
